#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from zun import objects
from zun.pci import stats
from zun.scheduler.filters import pci_passthrough_filter
from zun.scheduler.host_state import HostState
from zun.tests import base


class TestPCIPassthroughFilter(base.TestCase):

    def setUp(self):
        super(TestPCIPassthroughFilter, self).setUp()
        self.filt_cls = pci_passthrough_filter.PciPassthroughFilter()

    def test_pci_passthrough_pass(self):
        pci_stats_mock = mock.MagicMock()
        pci_stats_mock.support_requests.return_value = True
        request = objects.ContainerPCIRequest(
            count=1, spec=[{'vendor_id': '8086'}])
        requests = objects.ContainerPCIRequests(requests=[request])
        container = objects.Container(self.context)
        host = HostState('testhost')
        host.pci_stats = pci_stats_mock
        extra_spec = {'pci_requests': requests}
        self.assertTrue(self.filt_cls.host_passes(host, container, extra_spec))
        pci_stats_mock.support_requests.assert_called_once_with(
            requests.requests)

    def test_pci_passthrough_fail(self):
        pci_stats_mock = mock.MagicMock()
        pci_stats_mock.support_requests.return_value = False
        request = objects.ContainerPCIRequest(
            count=1, spec=[{'vendor_id': '8086'}])
        requests = objects.ContainerPCIRequests(requests=[request])
        container = objects.Container(self.context)
        host = HostState('testhost')
        host.pci_stats = pci_stats_mock
        extra_spec = {'pci_requests': requests}
        self.assertFalse(self.filt_cls.host_passes(host, container,
                                                   extra_spec))
        pci_stats_mock.support_requests.assert_called_once_with(
            requests.requests)

    def test_pci_passthrough_no_pci_request(self):
        container = objects.Container(self.context)
        host = HostState('testhost')
        extra_spec = {'pci_requests': None}
        self.assertTrue(self.filt_cls.host_passes(host, container, extra_spec))

    def test_pci_passthrough_empty_pci_request_obj(self):
        requests = objects.ContainerPCIRequests(requests=[])
        container = objects.Container(self.context)
        host = HostState('testhost')
        extra_spec = {'pci_requests': requests}
        self.assertTrue(self.filt_cls.host_passes(host, container, extra_spec))

    def test_pci_passthrough_no_pci_stats(self):
        request = objects.ContainerPCIRequest(
            count=1, spec=[{'vendor_id': '8086'}])
        requests = objects.ContainerPCIRequests(requests=[request])
        container = objects.Container(self.context)
        host = HostState('testhost')
        host.pci_stats = stats.PciDeviceStats()
        extra_spec = {'pci_requests': requests}
        self.assertFalse(self.filt_cls.host_passes(host, container,
                                                   extra_spec))

    def test_pci_passthrough_with_pci_stats_none(self):
        request = objects.ContainerPCIRequest(
            count=1, spec=[{'vendor_id': '8086'}])
        requests = objects.ContainerPCIRequests(requests=[request])
        container = objects.Container(self.context)
        host = HostState('testhost')
        host.pci_stats = None
        extra_spec = {'pci_requests': requests}
        self.assertFalse(self.filt_cls.host_passes(host, container,
                                                   extra_spec))



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Add cpu workdir ports hostname labels to container

Revision ID: 1192ba19a6e9
Revises: 63a08e32cc43
Create Date: 2016-08-20 09:56:38.902481

"""

# revision identifiers, used by Alembic.
revision = '1192ba19a6e9'
down_revision = '63a08e32cc43'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa

import zun


def upgrade():
    op.add_column('container',
                  sa.Column('cpu', sa.Float(),
                            nullable=True))
    op.add_column('container',
                  sa.Column('workdir', sa.String(length=255),
                            nullable=True))
    op.add_column('container',
                  sa.Column('ports',
                            zun.db.sqlalchemy.models.JSONEncodedList(),
                            nullable=True))
    op.add_column('container',
                  sa.Column('hostname', sa.String(length=255),
                            nullable=True))
    op.add_column('container',
                  sa.Column('labels',
                            zun.db.sqlalchemy.models.JSONEncodedDict(),
                            nullable=True))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from zun.common import context
from zun import objects
from zun.scheduler.filters import availability_zone_filter as az_filter
from zun.tests import base
from zun.tests.unit.scheduler import fakes


class TestAvailabilityZoneFilter(base.TestCase):

    def setUp(self):
        super(TestAvailabilityZoneFilter, self).setUp()
        self.context = context.RequestContext('fake_user', 'fake_project')

    def test_az_filter(self):
        self.assertIs(True,
                      self._test_az_filter(request_az='test-az',
                                           node_az='test-az'))
        self.assertIs(False,
                      self._test_az_filter(request_az='test-az',
                                           node_az='another-az'))

    def test_az_filter_default_az(self):
        cfg.CONF.set_override("default_availability_zone", "default-az")
        self.assertIs(True,
                      self._test_az_filter(request_az='default-az',
                                           node_az=None))
        self.assertIs(False,
                      self._test_az_filter(request_az='another-az',
                                           node_az=None))

    def test_az_filter_default_schedule_az(self):
        cfg.CONF.set_override("default_schedule_zone", "schedule-az")
        self.assertIs(True,
                      self._test_az_filter(request_az=None,
                                           node_az='schedule-az'))
        self.assertIs(False,
                      self._test_az_filter(request_az=None,
                                           node_az='another-az'))

    def test_az_filter_no_az_requested(self):
        self.assertIs(True,
                      self._test_az_filter(request_az=None,
                                           node_az=None))
        self.assertIs(True,
                      self._test_az_filter(request_az=None,
                                           node_az='any-az'))

    def _test_az_filter(self, request_az, node_az):
        filt_cls = az_filter.AvailabilityZoneFilter()
        container = objects.Container(self.context)
        service = objects.ZunService(self.context)
        service.availability_zone = node_az
        extra_spec = {}
        if request_az:
            extra_spec = {'availability_zone': request_az}
        host = fakes.FakeHostState('fake-host',
                                   {'service': service})
        return filt_cls.host_passes(host, container, extra_spec)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from zun.common import exception
import zun.conf
from zun.db import api as dbapi
from zun.tests.unit.db import base
from zun.tests.unit.db import utils

CONF = zun.conf.CONF


class DbVolumeTestCase(base.DbTestCase):

    def setUp(self):
        super(DbVolumeTestCase, self).setUp()

    def test_create_volume(self):
        utils.create_test_volume(context=self.context)

    def test_create_volume_already_exists(self):
        utils.create_test_volume(context=self.context,
                                 uuid='123')
        with self.assertRaisesRegex(exception.VolumeAlreadyExists,
                                    'A volume with UUID 123.*'):
            utils.create_test_volume(context=self.context,
                                     uuid='123')

    def test_get_volume_by_id(self):
        volume = utils.create_test_volume(context=self.context)
        res = dbapi.get_volume_by_id(self.context, volume.id)
        self.assertEqual(volume.id, res.id)
        self.assertEqual(volume.uuid, res.uuid)

    def test_get_volume_that_does_not_exist(self):
        self.assertRaises(exception.VolumeNotFound,
                          dbapi.get_volume_by_id,
                          self.context,
                          134251)

    def test_destroy_volume(self):
        volume = utils.create_test_volume(context=self.context)
        dbapi.destroy_volume(self.context, volume.id)
        self.assertRaises(exception.VolumeNotFound,
                          dbapi.get_volume_by_id,
                          self.context, volume.id)

    def test_destroy_volume_that_does_not_exist(self):
        self.assertRaises(exception.VolumeNotFound,
                          dbapi.destroy_volume, self.context,
                          134251)

    def test_update_volume(self):
        volume = utils.create_test_volume(context=self.context)
        old_conn_info = volume.connection_info
        new_conn_info = 'new-conn-info'
        self.assertNotEqual(old_conn_info, new_conn_info)

        res = dbapi.update_volume(self.context, volume.id,
                                  {'connection_info': new_conn_info})
        self.assertEqual(new_conn_info, res.connection_info)

    def test_update_volume_not_found(self):
        new_conn_info = 'new-conn-info'
        self.assertRaises(exception.VolumeNotFound,
                          dbapi.update_volume, self.context,
                          134251,
                          {'connection_info': new_conn_info})

    def test_update_volume_uuid(self):
        volume = utils.create_test_volume(context=self.context)
        self.assertRaises(exception.InvalidParameterValue,
                          dbapi.update_volume, self.context,
                          volume.id, {'uuid': ''})



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys

from oslo_log import log

from zun.api import app
from zun.common import profiler
from zun.common import service
import zun.conf


CONF = zun.conf.CONF
LOG = log.getLogger(__name__)


def init_application():
    # Initialize the oslo configuration library and logging
    service.prepare_service(sys.argv)
    profiler.setup('zun-api', CONF.host)

    LOG.debug("Configuration:")
    CONF.log_opt_values(LOG, log.DEBUG)

    return app.load_app()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import jsonschema
import six

from zun.common import exception
from zun.common.i18n import _


class SchemaValidator(object):
    """Resource reference validator class."""

    validator_org = jsonschema.Draft4Validator

    def __init__(self, schema, is_body=True):
        self.is_body = is_body
        validators = {
            'minimum': self._validate_minimum,
            'maximum': self._validate_maximum
        }
        validator_cls = jsonschema.validators.extend(self.validator_org,
                                                     validators)
        fc = jsonschema.FormatChecker()
        self.validator = validator_cls(schema, format_checker=fc)

    def validate(self, *args, **kwargs):
        try:
            self.validator.validate(*args, **kwargs)
        except jsonschema.ValidationError as ex:
            if len(ex.path) > 0:
                if self.is_body:
                    detail = _("Invalid input for field '%(path)s'."
                               "Value: '%(value)s'. %(message)s")
                else:
                    detail = _("Invalid input for query parameters "
                               "'%(path)s'. Value: '%(value)s'. %(message)s")
                detail = detail % {
                    'path': ex.path.pop(), 'value': ex.instance,
                    'message': six.text_type(ex)
                }
            else:
                detail = ex.message
            raise exception.SchemaValidationError(detail=detail)

    def _number_from_str(self, instance):
        if isinstance(instance, float) or isinstance(instance, int):
            return instance

        try:
            value = int(instance)
        except (ValueError, TypeError):
            try:
                value = float(instance)
            except (ValueError, TypeError):
                return None
        return value

    def _validate_minimum(self, validator, minimum, instance, schema):
        instance = self._number_from_str(instance)
        if instance is None:
            return
        return self.validator_org.VALIDATORS['minimum'](validator, minimum,
                                                        instance, schema)

    def _validate_maximum(self, validator, maximum, instance, schema):
        instance = self._number_from_str(instance)
        if instance is None:
            return
        return self.validator_org.VALIDATORS['maximum'](validator, maximum,
                                                        instance, schema)



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
import pecan

from zun.api.controllers import base
from zun.api.controllers import link
from zun.api.controllers.v1 import collection
from zun.api.controllers.v1.schemas import registries as schema
from zun.api.controllers.v1.views import registries_view as view
from zun.api import utils as api_utils
from zun.api import validation
from zun.common import exception
from zun.common.i18n import _
from zun.common.policies import registry as policies
from zun.common import policy
from zun.common import utils
import zun.conf
from zun import objects


CONF = zun.conf.CONF

LOG = logging.getLogger(__name__)

RESOURCE_NAME = 'registry'
COLLECTION_NAME = 'registries'


def check_policy_on_registry(registry, action):
    context = pecan.request.context
    policy.enforce(context, action, registry, action=action)


class RegistryCollection(collection.Collection):
    """API representation of a collection of registries."""

    fields = {
        COLLECTION_NAME,
        'next'
    }

    """A list containing registries objects"""

    def __init__(self, **kwargs):
        super(RegistryCollection, self).__init__(**kwargs)
        self._type = COLLECTION_NAME

    @staticmethod
    def convert_with_links(rpc_registries, limit, url=None,
                           **kwargs):
        context = pecan.request.context
        collection = RegistryCollection()
        collection.registries = \
            [view.format_registry(context, url, r.as_dict())
             for r in rpc_registries]
        collection.next = collection.get_next(limit, url=url, **kwargs)
        return collection


class RegistryItem(collection.Item):
    """API representation of an registry."""

    fields = {
        RESOURCE_NAME,
    }

    @staticmethod
    def render_response(rpc_registry):
        context = pecan.request.context
        url = pecan.request.host_url
        item = RegistryItem()
        item.registry = view.format_registry(context, url,
                                             rpc_registry.as_dict())
        return item


class RegistryController(base.Controller):
    """Controller for Registries."""

    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    def get_all(self, **kwargs):
        """Retrieve a list of registries.

        """
        context = pecan.request.context
        policy_action = policies.REGISTRY % 'get_all'
        policy.enforce(context, policy_action, action=policy_action)
        return self._get_registries_collection(**kwargs)

    def _get_registries_collection(self, **kwargs):
        context = pecan.request.context
        if utils.is_all_projects(kwargs):
            policy_action = policies.REGISTRY % 'get_all_all_projects'
            policy.enforce(context, policy_action, action=policy_action)
            context.all_projects = True
        kwargs.pop('all_projects', None)
        limit = api_utils.validate_limit(kwargs.pop('limit', None))
        sort_dir = api_utils.validate_sort_dir(kwargs.pop('sort_dir', 'asc'))
        sort_key = kwargs.pop('sort_key', 'id')
        resource_url = kwargs.pop('resource_url', None)

        registry_allowed_filters = ['name', 'domain', 'username',
                                    'project_id', 'user_id']
        filters = {}
        for filter_key in registry_allowed_filters:
            if filter_key in kwargs:
                policy_action = policies.REGISTRY % ('get_one:' + filter_key)
                context.can(policy_action, might_not_exist=True)
                filter_value = kwargs.pop(filter_key)
                filters[filter_key] = filter_value
        marker_obj = None
        marker = kwargs.pop('marker', None)
        if marker:
            marker_obj = objects.Registry.get_by_uuid(context, marker)
        if kwargs:
            unknown_params = [str(k) for k in kwargs]
            msg = _("Unknown parameters: %s") % ", ".join(unknown_params)
            raise exception.InvalidValue(msg)

        registries = objects.Registry.list(context,
                                           limit,
                                           marker_obj,
                                           sort_key,
                                           sort_dir,
                                           filters=filters)
        return RegistryCollection.convert_with_links(registries, limit,
                                                     url=resource_url,
                                                     sort_key=sort_key,
                                                     sort_dir=sort_dir)

    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    def get_one(self, registry_ident, **kwargs):
        """Retrieve information about the given registry.

        :param registry_ident: UUID or name of a registry.
        """
        context = pecan.request.context
        if context.is_admin:
            context.all_projects = True
        registry = utils.get_registry(registry_ident)
        policy_action = policies.REGISTRY % 'get_one'
        check_policy_on_registry(registry.as_dict(), policy_action)
        return RegistryItem.render_response(registry)

    @pecan.expose('json')
    @api_utils.enforce_content_types(['application/json'])
    @exception.wrap_pecan_controller_exception
    @validation.validated(schema.registry_create)
    def post(self, run=False, **registry_dict):
        context = pecan.request.context
        policy_action = policies.REGISTRY % 'create'
        policy.enforce(context, policy_action, action=policy_action)
        registry_dict = registry_dict.get(RESOURCE_NAME)
        registry_dict['project_id'] = context.project_id
        registry_dict['user_id'] = context.user_id
        new_registry = objects.Registry(context, **registry_dict)
        new_registry.create(context)
        # Set the HTTP Location Header
        pecan.response.location = link.build_url(COLLECTION_NAME,
                                                 new_registry.uuid)
        pecan.response.status = 201
        return RegistryItem.render_response(new_registry)

    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    @validation.validated(schema.registry_update)
    def patch(self, registry_ident, **registry_dict):
        """Update an existing registry.

        :param registry_ident: UUID or name of a registry.
        :param registry_dict: a json document to apply to this registry.
        """
        registry = utils.get_registry(registry_ident)
        context = pecan.request.context
        policy_action = policies.REGISTRY % 'update'
        check_policy_on_registry(registry.as_dict(), policy_action)
        registry_dict = registry_dict.get(RESOURCE_NAME)
        if 'name' in registry_dict:
            registry.name = registry_dict['name']
        if 'domain' in registry_dict:
            registry.domain = registry_dict['domain']
        if 'username' in registry_dict:
            registry.username = registry_dict['username']
        if 'password' in registry_dict:
            registry.password = registry_dict['password']
        registry.save(context)
        return RegistryItem.render_response(registry)

    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    def delete(self, registry_ident, **kwargs):
        """Delete a registry.

        :param registry_ident: UUID or Name of a registry.
        :param force: If True, allow to force delete the registry.
        """
        context = pecan.request.context
        if context.is_admin:
            context.all_projects = True
        registry = utils.get_registry(registry_ident)
        policy_action = policies.REGISTRY % 'delete'
        check_policy_on_registry(registry.as_dict(), policy_action)
        registry.destroy(context)
        pecan.response.status = 204



#    Copyright 2016 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from zun.common.utils import check_container_id
from zun.container import driver


class FakeDriver(driver.ContainerDriver):
    """Fake driver for testing."""

    def __init__(self):
        super(FakeDriver, self).__init__()

    def load_image(self, image, image_path=None):
        pass

    def inspect_image(self, image):
        pass

    def get_image(self, name):
        pass

    def delete_image(self, img_id):
        pass

    def images(self, repo, **kwargs):
        pass

    def pull_image(self, context, repo, tag, **kwargs):
        pass

    def create_image(self, context, image_name, image_driver):
        pass

    def upload_image_data(self, context, image, image_tag, image_data,
                          image_driver):
        pass

    def create(self, container):
        pass

    def delete(self, container, force):
        pass

    def list(self):
        pass

    def show(self, context, container):
        pass

    def create_capsule(self, context, capsule, **kwargs):
        pass

    def delete_capsule(self, context, capsule, **kwargs):
        pass

    @check_container_id
    def reboot(self, context, container):
        pass

    @check_container_id
    def stop(self, context, container):
        pass

    @check_container_id
    def start(self, context, container):
        pass

    @check_container_id
    def pause(self, context, container):
        pass

    @check_container_id
    def unpause(self, context, container):
        pass

    @check_container_id
    def show_logs(self, context, container):
        pass

    @check_container_id
    def execute(self, context, container, command):
        pass

    @check_container_id
    def kill(self, context, container, signal=None):
        pass

    @check_container_id
    def get_websocket_url(self, context, container):
        pass

    @check_container_id
    def resize(self, context, container, height, weight):
        pass

    def get_addresses(self, context, container):
        pass

    @check_container_id
    def update(self, context, container):
        pass

    @check_container_id
    def commit(self, context, container, repository, tag):
        pass

    def read_tar_image(self, image):
        return image.get('repo'), image.get('tag')

    def check_container_exist(self, context):
        pass

    def node_support_disk_quota(self):
        return True

    def get_host_default_base_size(self):
        return None



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import os
import types

from oslo_log import log as logging
from oslo_utils import fileutils
import six

from zun.common import exception
from zun.common.i18n import _
from zun.common import utils as common_utils
import zun.conf
from zun.image import driver
from zun.image.glance import utils

CONF = zun.conf.CONF

LOG = logging.getLogger(__name__)


class GlanceDriver(driver.ContainerImageDriver):
    def __init__(self):
        super(GlanceDriver, self).__init__()

    def _search_image_on_host(self, context, repo, tag):
        LOG.debug('Searching for image %s locally', repo)
        images_directory = CONF.glance.images_directory
        try:
            # TODO(mkrai): Change this to search image entry in zun db
            #              after the image endpoint is merged.
            image_meta = utils.find_image(context, repo, tag)
        except exception.ImageNotFound:
            return None
        if image_meta:
            out_path = os.path.join(images_directory,
                                    image_meta.id + '.tar')
            if os.path.isfile(out_path):
                return {
                    'image': repo,
                    'path': out_path,
                    'checksum': image_meta.checksum}
            else:
                return None

    def _verify_md5sum_for_image(self, image):
        image_path = image['path']
        image_checksum = image['checksum']
        md5sum = hashlib.md5()
        with open(image_path, 'rb') as fd:
            while True:
                # read 10MB of data each time
                data = fd.read(10 * 1024 * 1024)
                if not data:
                    break
                md5sum.update(data)
        md5sum = md5sum.hexdigest()
        if md5sum == image_checksum:
            return True
        return False

    def pull_image(self, context, repo, tag, image_pull_policy, registry):
        image_loaded = False
        image = self._search_image_on_host(context, repo, tag)

        if not common_utils.should_pull_image(image_pull_policy, bool(image)):
            if image:
                if self._verify_md5sum_for_image(image):
                    image_loaded = True
                    return image, image_loaded
            else:
                message = _('Image %s not present with pull policy of Never'
                            ) % repo
                raise exception.ImageNotFound(message)

        LOG.debug('Pulling image from glance %s', repo)
        try:
            image_meta = utils.find_image(context, repo, tag)
            LOG.debug('Image %s was found in glance, downloading now...', repo)
            image_chunks = utils.download_image_in_chunks(context,
                                                          image_meta.id)
        except exception.ImageNotFound:
            LOG.error('Image %s was not found in glance', repo)
            raise
        except Exception as e:
            msg = _('Cannot download image from glance: {0}')
            raise exception.ZunException(msg.format(e))
        try:
            images_directory = CONF.glance.images_directory
            fileutils.ensure_tree(images_directory)
            out_path = os.path.join(images_directory, image_meta.id + '.tar')
            with open(out_path, 'wb') as fd:
                for chunk in image_chunks:
                    fd.write(chunk)
        except Exception as e:
            msg = _('Error occurred while writing image: {0}')
            raise exception.ZunException(msg.format(e))
        LOG.debug('Image %(repo)s was downloaded to path : %(path)s',
                  {'repo': repo, 'path': out_path})
        image = {'image': image_meta.name, 'tags': image_meta.tags,
                 'path': out_path}
        return image, image_loaded

    def search_image(self, context, repo, tag, exact_match):
        LOG.debug('Searching image in glance %s', repo)
        try:
            return utils.find_images(context, repo, tag, exact_match)
        except Exception as e:
            raise exception.ZunException(six.text_type(e))

    def create_image(self, context, image_name):
        """Create an image."""
        LOG.debug('Creating a new image in glance %s', image_name)
        try:
            # Return a created image
            return utils.create_image(context, image_name)
        except Exception as e:
            raise exception.ZunException(six.text_type(e))

    def update_image(self, context, img_id, disk_format='qcow2',
                     container_format='docker', tag=None):
        """Update an image."""
        LOG.debug('Updating an image %s in glance', img_id)
        try:
            # NOTE(kiennt): Tags will be an empty list if no tag is defined.
            tags = [tag] if tag else []
            # Return the updated image
            return utils.update_image(context, img_id, disk_format,
                                      container_format, tags=tags)
        except Exception as e:
            raise exception.ZunException(six.text_type(e))

    def upload_image_data(self, context, img_id, data):
        """Upload an image."""
        LOG.debug('Uploading an image to glance %s', img_id)
        try:
            if isinstance(data, types.GeneratorType):
                # NOTE(kiennt): In Docker-py 3.1.0, get_image
                #               returns generator - related bugs [1].
                #               These lines makes image_data readable.
                # [1] https://bugs.launchpad.net/zun/+bug/1753080
                data = six.b('').join(data)
                data = six.BytesIO(data)

            return utils.upload_image_data(context, img_id, data)
        except Exception as e:
            raise exception.ZunException(six.text_type(e))

    def delete_committed_image(self, context, img_id):
        """Delete a committed image."""
        LOG.debug('Delete the committed image %s in glance', img_id)
        try:
            return utils.delete_image(context, img_id)
        except Exception as e:
            LOG.exception('Unknown exception occurred while deleting '
                          'image %s in glance: %s',
                          img_id,
                          six.text_type(e))
            raise exception.ZunException(six.text_type(e))

    def delete_image_tar(self, context, image):
        """Delete image tar file that pull from glance"""
        repo = image.split(':')[0]
        tag = image.split(':')[1]
        image = self._search_image_on_host(context, repo, tag)
        if image:
            tarfile = image.get('path')
            try:
                os.unlink(tarfile)
            except Exception as e:
                LOG.exception('Cannot delete tar file %s', tarfile)
                raise exception.ZunException(six.text_type(e))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import contextlib
import six
import sys
import tarfile

import docker
from docker import errors
from oslo_serialization import jsonutils
from oslo_utils import encodeutils

from zun.common import consts
from zun.common import exception
from zun.common.i18n import _
import zun.conf


CONF = zun.conf.CONF


@contextlib.contextmanager
def docker_client():
    client_kwargs = dict()
    if not CONF.docker.api_insecure:
        client_kwargs['ca_cert'] = CONF.docker.ca_file
        client_kwargs['client_key'] = CONF.docker.key_file
        client_kwargs['client_cert'] = CONF.docker.cert_file

    try:
        yield DockerHTTPClient(
            CONF.docker.api_url,
            CONF.docker.docker_remote_api_version,
            CONF.docker.default_timeout,
            **client_kwargs
        )
    except errors.APIError as e:
        desired_exc = exception.DockerError(error_msg=six.text_type(e))
        six.reraise(type(desired_exc), desired_exc, sys.exc_info()[2])


class DockerHTTPClient(docker.APIClient):
    def __init__(self, url=CONF.docker.api_url,
                 ver=CONF.docker.docker_remote_api_version,
                 timeout=CONF.docker.default_timeout,
                 ca_cert=None,
                 client_key=None,
                 client_cert=None):

        if ca_cert and client_key and client_cert:
            ssl_config = docker.tls.TLSConfig(
                client_cert=(client_cert, client_key),
                verify=ca_cert,
                assert_hostname=False,
            )
        else:
            ssl_config = False

        super(DockerHTTPClient, self).__init__(
            base_url=url,
            version=ver,
            timeout=timeout,
            tls=ssl_config
        )

    def list_containers(self):
        return self.containers(all=True, filters={'name': consts.NAME_PREFIX})

    def read_tar_image(self, image):
        image_path = image['path']
        with tarfile.open(image_path, 'r') as fil:
            fest = fil.extractfile('manifest.json')
            data = fest.read()
            data = jsonutils.loads(encodeutils.safe_decode(data))
            repo_tags = data[0]['RepoTags']
            if repo_tags:
                repo, tag = repo_tags[0].split(":")
                image['repo'], image['tag'] = repo, tag
            else:
                image_uuid = data[0]['Config'].split('.')[0]
                image['repo'], image['tag'] = image_uuid, ''

    def exec_resize(self, exec_id, height=None, width=None):
        # NOTE(hongbin): This is a temporary work-around for a docker issue
        # See: https://github.com/moby/moby/issues/35561
        try:
            super(DockerHTTPClient, self).exec_resize(
                exec_id, height=height, width=width)
        except errors.APIError as e:
            if "process not found for container" in str(e):
                raise exception.Invalid(_(
                    "no such exec instance: %s") % six.text_type(e))
            raise



# Copyright (c) 2010 OpenStack Foundation
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Scheduler base class that all Schedulers should inherit from
"""

import abc

import six

from zun.api import servicegroup
from zun import objects


@six.add_metaclass(abc.ABCMeta)
class Scheduler(object):
    """The base class that all Scheduler classes should inherit from."""

    def __init__(self):
        self.servicegroup_api = servicegroup.ServiceGroup()

    def hosts_up(self, context):
        """Return the list of hosts that have a running service."""

        services = objects.ZunService.list_by_binary(context, 'zun-compute')
        return [service.host
                for service in services
                if self.servicegroup_api.service_is_up(service)
                and not service.disabled]

    @abc.abstractmethod
    def select_destinations(self, context, containers, extra_specs,
                            alloc_reqs_by_rp_uuid, provider_summaries,
                            allocation_request_version=None):
        """Must override select_destinations method.

        :return: A list of dicts with 'host', 'nodename' and 'limits' as keys
            that satisfies the extra_specs and filter_properties.
        """
        raise NotImplementedError()



# Copyright 2012 OpenStack Foundation  # All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests For Loadable class handling.
"""

from zun.common import exception
from zun.tests import base
from zun.tests.unit.scheduler import fake_loadables


class LoadablesTestCase(base.TestCase):
    def setUp(self):
        super(LoadablesTestCase, self).setUp()
        self.fake_loader = fake_loadables.FakeLoader()
        # The name that we imported above for testing
        self.test_package = 'zun.tests.unit.scheduler.fake_loadables'

    def test_loader_init(self):
        self.assertEqual(self.fake_loader.package, self.test_package)
        # Test the path of the module
        ending_path = '/' + self.test_package.replace('.', '/')
        self.assertTrue(self.fake_loader.path.endswith(ending_path))
        self.assertEqual(self.fake_loader.loadable_cls_type,
                         fake_loadables.FakeLoadable)

    def _compare_classes(self, classes, expected):
        class_names = [cls.__name__ for cls in classes]
        self.assertEqual(set(class_names), set(expected))

    def test_get_all_classes(self):
        classes = self.fake_loader.get_all_classes()
        expected_class_names = ['FakeLoadableSubClass1',
                                'FakeLoadableSubClass2',
                                'FakeLoadableSubClass5',
                                'FakeLoadableSubClass6']
        self._compare_classes(classes, expected_class_names)

    def test_get_matching_classes(self):
        prefix = self.test_package
        test_classes = [prefix + '.fake_loadable1.FakeLoadableSubClass1',
                        prefix + '.fake_loadable2.FakeLoadableSubClass5']
        classes = self.fake_loader.get_matching_classes(test_classes)
        expected_class_names = ['FakeLoadableSubClass1',
                                'FakeLoadableSubClass5']
        self._compare_classes(classes, expected_class_names)

    def test_get_matching_classes_with_underscore(self):
        prefix = self.test_package
        test_classes = [prefix + '.fake_loadable1.FakeLoadableSubClass1',
                        prefix + '.fake_loadable2._FakeLoadableSubClass7']
        self.assertRaises(exception.ClassNotFound,
                          self.fake_loader.get_matching_classes,
                          test_classes)

    def test_get_matching_classes_with_wrong_type1(self):
        prefix = self.test_package
        test_classes = [prefix + '.fake_loadable1.FakeLoadableSubClass4',
                        prefix + '.fake_loadable2.FakeLoadableSubClass5']
        self.assertRaises(exception.ClassNotFound,
                          self.fake_loader.get_matching_classes,
                          test_classes)

    def test_get_matching_classes_with_wrong_type2(self):
        prefix = self.test_package
        test_classes = [prefix + '.fake_loadable1.FakeLoadableSubClass1',
                        prefix + '.fake_loadable2.FakeLoadableSubClass8']
        self.assertRaises(exception.ClassNotFound,
                          self.fake_loader.get_matching_classes,
                          test_classes)

    def test_get_matching_classes_with_one_function(self):
        prefix = self.test_package
        test_classes = [prefix + '.fake_loadable1.return_valid_classes',
                        prefix + '.fake_loadable2.FakeLoadableSubClass5']
        classes = self.fake_loader.get_matching_classes(test_classes)
        expected_class_names = ['FakeLoadableSubClass1',
                                'FakeLoadableSubClass2',
                                'FakeLoadableSubClass5']
        self._compare_classes(classes, expected_class_names)

    def test_get_matching_classes_with_two_functions(self):
        prefix = self.test_package
        test_classes = [prefix + '.fake_loadable1.return_valid_classes',
                        prefix + '.fake_loadable2.return_valid_class']
        classes = self.fake_loader.get_matching_classes(test_classes)
        expected_class_names = ['FakeLoadableSubClass1',
                                'FakeLoadableSubClass2',
                                'FakeLoadableSubClass6']
        self._compare_classes(classes, expected_class_names)

    def test_get_matching_classes_with_function_including_invalids(self):
        # When using a method, no checking is done on valid classes.
        prefix = self.test_package
        test_classes = [prefix + '.fake_loadable1.return_invalid_classes',
                        prefix + '.fake_loadable2.return_valid_class']
        classes = self.fake_loader.get_matching_classes(test_classes)
        expected_class_names = ['FakeLoadableSubClass1',
                                '_FakeLoadableSubClass3',
                                'FakeLoadableSubClass4',
                                'FakeLoadableSubClass6']
        self._compare_classes(classes, expected_class_names)



#    Copyright (c) 2018 National Engineering Laboratory of electronic
#    commerce and electronic payment
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

import zun.conf
from zun.scheduler import filters

LOG = logging.getLogger(__name__)
CONF = zun.conf.CONF


class CpuSetFilter(filters.BaseHostFilter):
    """Filter the host by cpu and memory request of cpuset"""

    run_filter_once_per_request = True

    def host_passes(self, host_state, container, extra_spec):
        if container.cpu_policy is None:
            container.cpu_policy = 'shared'
        if container.memory is None:
            container_memory = 0
        else:
            container_memory = int(container.memory)
        if container.cpu_policy == 'dedicated':
            if host_state.enable_cpu_pinning:
                for numa_node in host_state.numa_topology.nodes:
                    if len(numa_node.cpuset) - len(
                            numa_node.pinned_cpus) >= container.cpu and \
                            numa_node.mem_available >= container_memory:
                        host_state.limits['cpuset'] = {
                            'node': numa_node.id,
                            'cpuset_cpu': numa_node.cpuset,
                            'cpuset_cpu_pinned': numa_node.pinned_cpus,
                            'cpuset_mem': numa_node.mem_available
                        }
                        return True
                return False
            else:
                return False
        if container.cpu_policy == 'shared':
            if CONF.compute.enable_cpu_pinning:
                return False
            else:
                return True



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Zun object test utilities."""


from zun import objects
from zun.objects.numa import NUMATopology
from zun.tests.unit.db import utils as db_utils


def create_test_container(context, **kwargs):
    """Create and return a test container object.

    Create a container in the DB and return a container object with
    appropriate attributes.
    """
    container = get_test_container(context, **kwargs)
    container.create(context)
    return container


def get_test_container(context, **kwargs):
    """Return a test container object with appropriate attributes.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    db_container = db_utils.get_test_container(**kwargs)
    container = objects.Container(context)
    for key in db_container:
        setattr(container, key, db_container[key])
    return container


def create_test_registry(context, **kwargs):
    """Create and return a test registry object.

    Create a registry in the DB and return a registry object with
    appropriate attributes.
    """
    registry = get_test_registry(context, **kwargs)
    registry.create(context)
    return registry


def get_test_registry(context, **kwargs):
    """Return a test registry object with appropriate attributes.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    db_registry = db_utils.get_test_registry(**kwargs)
    registry = objects.Registry(context)
    for key in db_registry:
        setattr(registry, key, db_registry[key])
    return registry


def get_test_compute_node(context, **kwargs):
    """Return a test compute node object with appropriate attributes.

    NOTE: The object leaves the attributes marked as changed, such
    that a create() could be used to commit it to the DB.
    """
    db_compute_node = db_utils.get_test_compute_node(**kwargs)
    compute_node = objects.ComputeNode(context)
    for key in db_compute_node:
        if key == 'numa_topology':
            numa_obj = NUMATopology._from_dict(db_compute_node[key])
            compute_node.numa_topology = numa_obj
        else:
            setattr(compute_node, key, db_compute_node[key])
    return compute_node



# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_concurrency import processutils
from oslo_log import log

from nova import exception
from nova import utils

from nova.i18n import _

LOG = log.getLogger(__name__)


def teardown_network(container_id):
    try:
        output, err = utils.execute('ip', '-o', 'netns', 'list')
        for line in output.split('\n'):
            if container_id == line.strip():
                utils.execute('ip', 'netns', 'delete', container_id,
                              run_as_root=True)
                break
    except processutils.ProcessExecutionError:
        LOG.warning('Cannot remove network namespace, netns id: %s',
                    container_id)


def find_fixed_ips(instance, network_info):
    ips = []
    for subnet in network_info['subnets']:
        netmask = subnet['cidr'].split('/')[1]
        for ip in subnet['ips']:
            if ip['type'] == 'fixed' and ip['address']:
                ips.append(ip['address'] + "/" + netmask)
    if not ips:
        raise exception.InstanceDeployFailure(_('Cannot find fixed ip'),
                                              instance_id=instance['uuid'])
    return ips


def find_gateways(instance, network_info):
    gateways = []
    for subnet in network_info['subnets']:
        gateways.append(subnet['gateway']['address'])
    if not gateways:
        raise exception.InstanceDeployFailure(_('Cannot find gateway'),
                                              instance_id=instance['uuid'])
    return gateways


# NOTE(arosen) - this method should be removed after it's moved into the
# linux_net code in nova.
def get_ovs_interfaceid(vif):
    return vif.get('ovs_interfaceid') or vif['id']



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_versionedobjects import fields

from zun.common import consts
from zun.common import exception
from zun.common.i18n import _
from zun.db import api as dbapi
from zun.objects import base
from zun.objects import exec_instance as exec_inst
from zun.objects import fields as z_fields
from zun.objects import pci_device
from zun.objects import registry


LOG = logging.getLogger(__name__)


CONTAINER_OPTIONAL_ATTRS = ["pci_devices", "exec_instances", "registry"]


@base.ZunObjectRegistry.register
class Cpuset(base.ZunObject):
    VERSION = '1.0'

    fields = {
        'cpuset_cpus': fields.SetOfIntegersField(nullable=True),
        'cpuset_mems': fields.SetOfIntegersField(nullable=True),
    }

    def _to_dict(self):
        return {
            'cpuset_cpus': self.cpuset_cpus,
            'cpuset_mems': self.cpuset_mems
        }

    @classmethod
    def _from_dict(cls, data_dict):
        if not data_dict:
            obj = cls(cpuset_cpus=None, cpuset_mems=None)
        else:
            cpuset_cpus = data_dict.get('cpuset_cpus')
            cpuset_mems = data_dict.get('cpuset_mems')
            obj = cls(cpuset_cpus=cpuset_cpus, cpuset_mems=cpuset_mems)
        obj.obj_reset_changes()
        return obj


class ContainerBase(base.ZunPersistentObject, base.ZunObject):

    fields = {
        'id': fields.IntegerField(),
        'container_id': fields.StringField(nullable=True),
        'uuid': fields.UUIDField(nullable=True),
        'name': fields.StringField(nullable=True),
        'project_id': fields.StringField(nullable=True),
        'user_id': fields.StringField(nullable=True),
        'image': fields.StringField(nullable=True),
        'cpu': fields.FloatField(nullable=True),
        'cpu_policy': fields.StringField(nullable=True),
        'cpuset': fields.ObjectField("Cpuset", nullable=True),
        'memory': fields.StringField(nullable=True),
        'command': fields.ListOfStringsField(nullable=True),
        'status': z_fields.ContainerStatusField(nullable=True),
        'status_reason': fields.StringField(nullable=True),
        'task_state': z_fields.TaskStateField(nullable=True),
        'environment': fields.DictOfStringsField(nullable=True),
        'workdir': fields.StringField(nullable=True),
        'auto_remove': fields.BooleanField(nullable=True),
        'ports': z_fields.ListOfIntegersField(nullable=True),
        'hostname': fields.StringField(nullable=True),
        'labels': fields.DictOfStringsField(nullable=True),
        'meta': fields.DictOfStringsField(nullable=True),
        'addresses': z_fields.JsonField(nullable=True),
        'image_pull_policy': fields.StringField(nullable=True),
        'host': fields.StringField(nullable=True),
        'restart_policy': fields.DictOfStringsField(nullable=True),
        'status_detail': fields.StringField(nullable=True),
        'interactive': fields.BooleanField(nullable=True),
        'tty': fields.BooleanField(nullable=True),
        'image_driver': fields.StringField(nullable=True),
        'websocket_url': fields.StringField(nullable=True),
        'websocket_token': fields.StringField(nullable=True),
        'security_groups': fields.ListOfStringsField(nullable=True),
        'runtime': fields.StringField(nullable=True),
        'pci_devices': fields.ListOfObjectsField('PciDevice',
                                                 nullable=True),
        'disk': fields.IntegerField(nullable=True),
        'auto_heal': fields.BooleanField(nullable=True),
        'started_at': fields.DateTimeField(tzinfo_aware=False, nullable=True),
        'exposed_ports': z_fields.JsonField(nullable=True),
        'exec_instances': fields.ListOfObjectsField('ExecInstance',
                                                    nullable=True),
        'privileged': fields.BooleanField(nullable=True),
        'healthcheck': z_fields.JsonField(nullable=True),
        'registry_id': fields.IntegerField(nullable=True),
        'registry': fields.ObjectField("Registry", nullable=True),
    }

    # should be redefined in subclasses
    container_type = None

    @staticmethod
    def _from_db_object(container, db_container):
        """Converts a database entity to a formal object."""
        for field in container.fields:
            if field in ['pci_devices', 'exec_instances', 'registry',
                         'containers', 'init_containers']:
                continue
            if field == 'cpuset':
                container.cpuset = Cpuset._from_dict(
                    db_container['cpuset'])
                continue
            setattr(container, field, db_container[field])

        container.obj_reset_changes()
        return container

    @staticmethod
    def _from_db_object_list(db_objects, cls, context):
        """Converts a list of database entities to a list of formal objects."""
        return [cls._from_db_object(cls(context), obj)
                for obj in db_objects]

    @base.remotable_classmethod
    def get_by_uuid(cls, context, uuid):
        """Find a container based on uuid and return a :class:`Container` object.

        :param uuid: the uuid of a container.
        :param context: Security context
        :returns: a :class:`Container` object.
        """
        db_container = dbapi.get_container_by_uuid(context, cls.container_type,
                                                   uuid)
        container = cls._from_db_object(cls(context), db_container)
        return container

    @base.remotable_classmethod
    def get_by_name(cls, context, name):
        """Find a container based on name and return a Container object.

        :param name: the logical name of a container.
        :param context: Security context
        :returns: a :class:`Container` object.
        """
        db_container = dbapi.get_container_by_name(context, cls.container_type,
                                                   name)
        container = cls._from_db_object(cls(context), db_container)
        return container

    @staticmethod
    def get_container_any_type(context, uuid):
        """Find a container of any type based on uuid.

        :param uuid: the uuid of a container.
        :param context: Security context
        :returns: a :class:`ContainerBase` object.
        """
        db_container = dbapi.get_container_by_uuid(context, consts.TYPE_ANY,
                                                   uuid)
        type = db_container['container_type']
        if type == consts.TYPE_CONTAINER:
            container_cls = Container
        elif type == consts.TYPE_CAPSULE:
            container_cls = Capsule
        elif type == consts.TYPE_CAPSULE_CONTAINER:
            container_cls = CapsuleContainer
        elif type == consts.TYPE_CAPSULE_INIT_CONTAINER:
            container_cls = CapsuleInitContainer
        else:
            raise exception.ZunException(_('Unknown container type: %s'), type)

        obj = container_cls(context)
        container = container_cls._from_db_object(obj, db_container)
        return container

    @base.remotable_classmethod
    def list(cls, context, limit=None, marker=None,
             sort_key=None, sort_dir=None, filters=None):
        """Return a list of Container objects.

        :param context: Security context.
        :param limit: maximum number of resources to return in a single result.
        :param marker: pagination marker for large data sets.
        :param sort_key: column to sort results by.
        :param sort_dir: direction to sort. "asc" or "desc".
        :param filters: filters when list containers, the filter name could be
                        'name', 'image', 'project_id', 'user_id', 'memory'.
                        For example, filters={'image': 'nginx'}
        :returns: a list of :class:`Container` object.

        """
        db_containers = dbapi.list_containers(
            context, cls.container_type, limit=limit, marker=marker,
            sort_key=sort_key, sort_dir=sort_dir, filters=filters)
        return cls._from_db_object_list(db_containers, cls, context)

    @base.remotable_classmethod
    def list_by_host(cls, context, host):
        """Return a list of Container objects by host.

        :param context: Security context.
        :param host: A compute host.
        :returns: a list of :class:`Container` object.

        """
        db_containers = dbapi.list_containers(context, cls.container_type,
                                              filters={'host': host})
        return cls._from_db_object_list(db_containers, cls, context)

    @base.remotable
    def create(self, context):
        """Create a Container record in the DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: Container(context)

        """
        values = self.obj_get_changes()
        cpuset_obj = values.pop('cpuset', None)
        if cpuset_obj is not None:
            values['cpuset'] = cpuset_obj._to_dict()
        values['container_type'] = self.container_type
        db_container = dbapi.create_container(context, values)
        self._from_db_object(self, db_container)

    @base.remotable
    def destroy(self, context=None):
        """Delete the Container from the DB.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: Container(context)
        """
        dbapi.destroy_container(context, self.container_type, self.uuid)
        self.obj_reset_changes()

    @base.remotable
    def save(self, context=None):
        """Save updates to this Container.

        Updates will be made column by column based on the result
        of self.what_changed().

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: Container(context)
        """
        updates = self.obj_get_changes()
        cpuset_obj = updates.pop('cpuset', None)
        if cpuset_obj is not None:
            updates['cpuset'] = cpuset_obj._to_dict()
        dbapi.update_container(context, self.container_type, self.uuid,
                               updates)

        self.obj_reset_changes()

    @base.remotable
    def refresh(self, context=None):
        """Loads updates for this Container.

        Loads a container with the same uuid from the database and
        checks for updated attributes. Updates are applied from
        the loaded container column by column, if there are any updates.

        :param context: Security context. NOTE: This should only
                        be used internally by the indirection_api.
                        Unfortunately, RPC requires context as the first
                        argument, even though we don't use it.
                        A context should be set when instantiating the
                        object, e.g.: Container(context)
        """
        current = self.__class__.get_by_uuid(self._context, uuid=self.uuid)
        for field in self.fields:
            if self.obj_attr_is_set(field) and \
               getattr(self, field) != getattr(current, field):
                setattr(self, field, getattr(current, field))

    def obj_load_attr(self, attrname):
        if attrname not in CONTAINER_OPTIONAL_ATTRS:
            raise exception.ObjectActionError(
                action='obj_load_attr',
                reason=_('attribute %s not lazy-loadable') % attrname)

        if not self._context:
            raise exception.OrphanedObjectError(method='obj_load_attr',
                                                objtype=self.obj_name())

        LOG.debug("Lazy-loading '%(attr)s' on %(name)s uuid %(uuid)s",
                  {'attr': attrname,
                   'name': self.obj_name(),
                   'uuid': self.uuid,
                   })

        # NOTE(danms): We handle some fields differently here so that we
        # can be more efficient
        if attrname == 'pci_devices':
            self._load_pci_devices()

        if attrname == 'exec_instances':
            self._load_exec_instances()

        if attrname == 'registry':
            self._load_registry()

        self.obj_reset_changes([attrname])

    def _load_pci_devices(self):
        self.pci_devices = pci_device.PciDevice.list_by_container_uuid(
            self._context, self.uuid)

    def _load_exec_instances(self):
        self.exec_instances = exec_inst.ExecInstance.list_by_container_id(
            self._context, self.id)

    def _load_registry(self):
        self.registry = None
        if self.registry_id:
            self.registry = registry.Registry.get_by_id(
                self._context, self.registry_id)

    @base.remotable_classmethod
    def get_count(cls, context, project_id, flag):
        """Get the counts of Container objects in the database.

        :param context: The request context for database access.
        :param project_id: The project_id to count across.
        :param flag: The name of resource, one of the following options:
                     - containers: Count the number of containers owned by the
                     project.
                     - memory: The sum of containers's memory.
                     - cpu: The sum of container's cpu.
                     - disk: The sum of container's disk size.
        """
        usage = dbapi.count_usage(context, cls.container_type, project_id,
                                  flag)[0] or 0
        return usage


@base.ZunObjectRegistry.register
class Container(ContainerBase):
    # Version 1.0: Initial version
    # Version 1.1: Add container_id column
    # Version 1.2: Add memory column
    # Version 1.3: Add task_state column
    # Version 1.4: Add cpu, workdir, ports, hostname and labels columns
    # Version 1.5: Add meta column
    # Version 1.6: Add addresses column
    # Version 1.7: Add host column
    # Version 1.8: Add restart_policy
    # Version 1.9: Add status_detail column
    # Version 1.10: Add tty, stdin_open
    # Version 1.11: Add image_driver
    # Version 1.12: Add 'Created' to ContainerStatus
    # Version 1.13: Add more task states for container
    # Version 1.14: Add method 'list_by_host'
    # Version 1.15: Combine tty and stdin_open
    # Version 1.16: Add websocket_url and token
    # Version 1.17: Add security_groups
    # Version 1.18: Add auto_remove
    # Version 1.19: Add runtime column
    # Version 1.20: Change runtime to String type
    # Version 1.21: Add pci_device attribute
    # Version 1.22: Add 'Deleting' to ContainerStatus
    # Version 1.23: Add the missing 'pci_devices' attribute
    # Version 1.24: Add the storage_opt attribute
    # Version 1.25: Change TaskStateField definition
    # Version 1.26:  Add auto_heal
    # Version 1.27: Make auto_heal field nullable
    # Version 1.28: Add 'Dead' to ContainerStatus
    # Version 1.29: Add 'Restarting' to ContainerStatus
    # Version 1.30: Add capsule_id attribute
    # Version 1.31: Add 'started_at' attribute
    # Version 1.32: Add 'exec_instances' attribute
    # Version 1.33: Change 'command' to List type
    # Version 1.34: Add privileged to container
    # Version 1.35: Add 'healthcheck' attribute
    # Version 1.36: Add 'get_count' method
    # Version 1.37: Add 'exposed_ports' attribute
    # Version 1.38: Add 'cpuset' attribute
    # Version 1.39: Add 'register' and 'registry_id' attributes
    # Version 1.40: Add 'tty' attributes
    VERSION = '1.40'

    container_type = consts.TYPE_CONTAINER


@base.ZunObjectRegistry.register
class Capsule(ContainerBase):
    # Version 1.0: Initial version
    # Version 1.1: Add 'tty' attributes
    VERSION = '1.1'

    container_type = consts.TYPE_CAPSULE

    fields = {
        'containers': fields.ListOfObjectsField('CapsuleContainer',
                                                nullable=True),
        'init_containers': fields.ListOfObjectsField('CapsuleInitContainer',
                                                     nullable=True),
    }

    def as_dict(self):
        capsule_dict = super(Capsule, self).as_dict()
        capsule_dict['containers'] = [c.as_dict() for c in self.containers]
        capsule_dict['init_containers'] = [c.as_dict()
                                           for c in self.init_containers]
        return capsule_dict

    def obj_load_attr(self, attrname):
        if attrname == 'containers':
            self._load_capsule_containers()
            self.obj_reset_changes([attrname])
        elif attrname == 'init_containers':
            self._load_capsule_init_containers()
            self.obj_reset_changes([attrname])
        else:
            super(Capsule, self).obj_load_attr(attrname)

    def _load_capsule_containers(self):
        self.containers = CapsuleContainer.list_by_capsule_id(
            self._context, self.id)

    def _load_capsule_init_containers(self):
        self.init_containers = CapsuleInitContainer.list_by_capsule_id(
            self._context, self.id)


@base.ZunObjectRegistry.register
class CapsuleContainer(ContainerBase):
    # Version 1.0: Initial version
    # Version 1.1: Add 'tty' attributes
    VERSION = '1.1'

    container_type = consts.TYPE_CAPSULE_CONTAINER

    fields = {
        'capsule_id': fields.IntegerField(nullable=False),
    }

    @base.remotable_classmethod
    def list_by_capsule_id(cls, context, capsule_id):
        """Return a list of Container objects by capsule_id.

        :param context: Security context.
        :param host: A capsule id.
        :returns: a list of :class:`Container` object.

        """
        db_containers = dbapi.list_containers(
            context, cls.container_type, filters={'capsule_id': capsule_id})
        return Container._from_db_object_list(db_containers, cls, context)


@base.ZunObjectRegistry.register
class CapsuleInitContainer(ContainerBase):
    # Version 1.0: Initial version
    # Version 1.1: Add 'tty' attributes
    VERSION = '1.1'

    container_type = consts.TYPE_CAPSULE_INIT_CONTAINER

    fields = {
        'capsule_id': fields.IntegerField(nullable=False),
    }

    @base.remotable_classmethod
    def list_by_capsule_id(cls, context, capsule_id):
        """Return a list of Container objects by capsule_id.

        :param context: Security context.
        :param host: A capsule id.
        :returns: a list of :class:`Container` object.

        """
        db_containers = dbapi.list_containers(
            context, cls.container_type, filters={'capsule_id': capsule_id})
        return Container._from_db_object_list(db_containers, cls, context)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add ondelete to container_actions_events foreign key

Revision ID: 50829990c965
Revises: fb9ad4a050f8
Create Date: 2018-02-02 03:58:14.384716

"""

# revision identifiers, used by Alembic.
revision = '50829990c965'
down_revision = 'fb9ad4a050f8'
branch_labels = None
depends_on = None

from alembic import op
from sqlalchemy.engine.reflection import Inspector as insp

CONTAINER_ACTIONS_EVENTS = 'container_actions_events'
CONTAINER_ACTIONS = 'container_actions'


def upgrade():
    bind = op.get_bind()

    inspector = insp.from_engine(bind)
    foreign_keys = inspector.get_foreign_keys(CONTAINER_ACTIONS_EVENTS)

    for foreign_key in foreign_keys:
        if foreign_key.get('referred_table') == CONTAINER_ACTIONS:
            op.drop_constraint(foreign_key.get('name'),
                               CONTAINER_ACTIONS_EVENTS,
                               type_="foreignkey")
            op.create_foreign_key(
                None, CONTAINER_ACTIONS_EVENTS, CONTAINER_ACTIONS,
                ['action_id'], ['id'], ondelete='CASCADE')



# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from zun.scheduler import filters

LOG = logging.getLogger(__name__)


class RamFilter(filters.BaseHostFilter):
    """Filter the containers by memory request"""

    run_filter_once_per_request = True

    def host_passes(self, host_state, container, extra_spec):
        if not container.memory:
            return True

        request_ram = int(container.memory)
        usable_ram = host_state.mem_total - host_state.mem_used
        if usable_ram < request_ram:
            LOG.debug("%(host_state)s does not have %(request_ram)d "
                      "usable memory, it only has %(usable_ram)d usable "
                      "memory.",
                      {'host_state': host_state,
                       'request_ram': request_ram,
                       'usable_ram': usable_ram})
            return False
        host_state.limits['memory'] = host_state.mem_total
        return True



# Copyright 2016 Intel.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import zun.conf
from zun.image.docker import driver as docker_driver
from zun.image import driver
from zun.image.glance import driver as glance_driver
from zun.tests import base

CONF = zun.conf.CONF


class TestDriver(base.BaseTestCase):
    def setUp(self):
        super(TestDriver, self).setUp()

    def test_load_image_driver_failure(self):
        CONF.set_override('default_image_driver', None)
        self.assertRaises(SystemExit, driver.load_image_driver)
        self.assertRaises(SystemExit, driver.load_image_driver,
                          'UnknownDriver')

    def test_load_image_driver(self):
        image_driver = driver.load_image_driver()
        self.assertIsInstance(image_driver, docker_driver.DockerDriver)

        CONF.set_override('images_directory', None, group='glance')
        image_driver = driver.load_image_driver('glance')
        self.assertIsInstance(image_driver, glance_driver.GlanceDriver)



# Copyright (c) 2017 Intel, Inc.
# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from oslo_config import cfg
from oslo_log import log as logging
import six

from zun.common import exception
from zun.objects import fields
from zun.objects import pci_device_pool
from zun.pci import utils
from zun.pci import whitelist


CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class PciDeviceStats(object):

    """PCI devices summary information.

    According to the PCI SR-IOV spec, a PCI physical function can have up to
    256 PCI virtual functions, thus the number of assignable PCI functions in
    a cloud can be big. The scheduler needs to know all device availability
    information in order to determine which compute hosts can support a PCI
    request. Passing individual virtual device information to the scheduler
    does not scale, so we provide summary information.

    Usually the virtual functions provided by a host PCI device have the same
    value for most properties, like vendor_id, product_id and class type.
    The PCI stats class summarizes this information for the scheduler.

    The pci stats information is maintained exclusively by compute node
    resource tracker and updated to database. The scheduler fetches the
    information and selects the compute node accordingly. If a compute
    node is selected, the resource tracker allocates the devices to the
    instance and updates the pci stats information.

    This summary information will be helpful for cloud management also.
    """

    pool_keys = ['product_id', 'vendor_id', 'numa_node', 'dev_type']

    def __init__(self, stats=None, dev_filter=None):
        super(PciDeviceStats, self).__init__()
        # NOTE(sbauza): Stats are a PCIDevicePoolList object
        self.pools = [pci_pool.to_dict()
                      for pci_pool in stats] if stats else []
        self.pools.sort(key=lambda item: len(item))
        self.dev_filter = dev_filter or whitelist.Whitelist(
            CONF.pci.passthrough_whitelist)

    def _equal_properties(self, dev, entry, matching_keys):
        return all(dev.get(prop) == entry.get(prop)
                   for prop in matching_keys)

    def _find_pool(self, dev_pool):
        """Return the first pool that matches dev."""
        for pool in self.pools:
            pool_keys = pool.copy()
            del pool_keys['count']
            del pool_keys['devices']
            equal = self._equal_properties(dev_pool, pool_keys,
                                           dev_pool.keys())
            if (len(pool_keys.keys()) == len(dev_pool.keys()) and equal):
                return pool

    def _create_pool_keys_from_dev(self, dev):
        """create a stats pool dict that this dev is supposed to be part of

        Note that this pool dict contains the stats pool's keys and their
        values. 'count' and 'devices' are not included.
        """
        # Don't add a device that doesn't have a matching device spec.
        # This can happen during initial sync up with the controller
        devspec = self.dev_filter.get_devspec(dev)
        if not devspec:
            return
        tags = devspec.get_tags()
        pool = {k: getattr(dev, k) for k in self.pool_keys}
        if tags:
            pool.update(tags)
        return pool

    def add_device(self, dev):
        """Add a device to its matching pool."""
        dev_pool = self._create_pool_keys_from_dev(dev)
        if dev_pool:
            pool = self._find_pool(dev_pool)
            if not pool:
                dev_pool['count'] = 0
                dev_pool['devices'] = []
                self.pools.append(dev_pool)
                self.pools.sort(key=lambda item: len(item))
                pool = dev_pool
            pool['count'] += 1
            pool['devices'].append(dev)

    @staticmethod
    def _decrease_pool_count(pool_list, pool, count=1):
        """Decrement pool's size by count.

        If pool becomes empty, remove pool from pool_list.
        """
        if pool['count'] > count:
            pool['count'] -= count
            count = 0
        else:
            count -= pool['count']
            pool_list.remove(pool)
        return count

    def remove_device(self, dev):
        """Remove one device from the first pool that it matches."""
        dev_pool = self._create_pool_keys_from_dev(dev)
        if dev_pool:
            pool = self._find_pool(dev_pool)
            if not pool:
                raise exception.PciDevicePoolEmpty(
                    compute_node_uuid=dev.compute_node_uuid,
                    address=dev.address)
            pool['devices'].remove(dev)
            self._decrease_pool_count(self.pools, pool)

    def get_free_devs(self):
        free_devs = []
        for pool in self.pools:
            free_devs.extend(pool['devices'])
        return free_devs

    def consume_requests(self, pci_requests, numa_cells=None):
        alloc_devices = []
        for request in pci_requests:
            count = request.count
            spec = request.spec
            # For now, keep the same algorithm as during scheduling:
            # a spec may be able to match multiple pools.
            pools = self._filter_pools_for_spec(self.pools, spec)
            if numa_cells:
                pools = self._filter_pools_for_numa_cells(pools, numa_cells)
            pools = self._filter_non_requested_pfs(request, pools)
            # Failed to allocate the required number of devices
            # Return the devices already allocated back to their pools
            if sum([pool['count'] for pool in pools]) < count:
                LOG.error("Failed to allocate PCI devices for container."
                          " Unassigning devices back to pools."
                          " This should not happen, since the scheduler"
                          " should have accurate information, and allocation"
                          " during claims is controlled via a hold"
                          " on the compute node semaphore")
                for d in range(len(alloc_devices)):
                    self.add_device(alloc_devices.pop())
                return None
            for pool in pools:
                if pool['count'] >= count:
                    num_alloc = count
                else:
                    num_alloc = pool['count']
                count -= num_alloc
                pool['count'] -= num_alloc
                for d in range(num_alloc):
                    pci_dev = pool['devices'].pop()
                    self._handle_device_dependents(pci_dev)
                    pci_dev.request_id = request.request_id
                    alloc_devices.append(pci_dev)
                if count == 0:
                    break
        return alloc_devices

    def _handle_device_dependents(self, pci_dev):
        """Remove device dependents or a parent from pools.

        In case the device is a PF, all of it's dependent VFs should
        be removed from pools count, if these are present.
        When the device is a VF, it's parent PF pool count should be
        decreased, unless it is no longer in a pool.
        """
        if pci_dev.dev_type == fields.PciDeviceType.SRIOV_PF:
            vfs_list = pci_dev.child_devices
            if vfs_list:
                for vf in vfs_list:
                    self.remove_device(vf)
        elif pci_dev.dev_type == fields.PciDeviceType.SRIOV_VF:
            try:
                parent = pci_dev.parent_device
                # Make sure not to decrease PF pool count if this parent has
                # been already removed from pools
                if parent in self.get_free_devs():
                    self.remove_device(parent)
            except exception.PciDeviceNotFound:
                return

    @staticmethod
    def _filter_pools_for_spec(pools, request_specs):
        return [pool for pool in pools
                if utils.pci_device_prop_match(pool, request_specs)]

    @staticmethod
    def _filter_pools_for_numa_cells(pools, numa_cells):
        # Some systems don't report numa node info for pci devices, in
        # that case None is reported in pci_device.numa_node, by adding None
        # to numa_cells we allow assigning those devices to instances with
        # numa topology
        numa_cells = [None] + [cell.id for cell in numa_cells]
        # filter out pools which numa_node is not included in numa_cells
        return [pool for pool in pools if any(utils.pci_device_prop_match(
                pool, [{'numa_node': cell}]) for cell in numa_cells)]

    def _filter_non_requested_pfs(self, request, matching_pools):
        # Remove SRIOV_PFs from pools, unless it has been explicitly requested
        # This is especially needed in cases where PFs and VFs has the same
        # product_id.
        if all(spec.get('dev_type') != fields.PciDeviceType.SRIOV_PF for
               spec in request.spec):
            matching_pools = self._filter_pools_for_pfs(matching_pools)
        return matching_pools

    @staticmethod
    def _filter_pools_for_pfs(pools):
        return [pool for pool in pools
                if not pool.get('dev_type') == fields.PciDeviceType.SRIOV_PF]

    def _apply_request(self, pools, request, numa_cells=None):
        # NOTE(vladikr): This code maybe open to race conditions.
        # Two concurrent requests may succeed when called support_requests
        # because this method does not remove related devices from the pools
        count = request.count
        matching_pools = self._filter_pools_for_spec(pools, request.spec)
        if numa_cells:
            matching_pools = self._filter_pools_for_numa_cells(matching_pools,
                                                               numa_cells)
        matching_pools = self._filter_non_requested_pfs(request,
                                                        matching_pools)
        if sum([pool['count'] for pool in matching_pools]) < count:
            return False
        else:
            for pool in matching_pools:
                count = self._decrease_pool_count(pools, pool, count)
                if not count:
                    break
        return True

    def support_requests(self, requests, numa_cells=None):
        """Check if the pci requests can be met.

        Scheduler checks compute node's PCI stats to decide if a
        container can be scheduled into the node. Support does not
        mean real allocation.
        If numa_cells is provided then only devices contained in
        those nodes are considered.
        """
        # note (yjiang5): this function has high possibility to fail,
        # so no exception should be triggered for performance reason.
        pools = copy.deepcopy(self.pools)
        return all([self._apply_request(pools, r, numa_cells)
                   for r in requests])

    def apply_requests(self, requests, numa_cells=None):
        """Apply PCI requests to the PCI stats.

        This is used in container creation, when the scheduler has to
        maintain how the resources are consumed by the container.
        If numa_cells is provided then only devices contained in
        those nodes are considered.
        """
        if not all([self._apply_request(self.pools, r, numa_cells)
                   for r in requests]):
            raise exception.PciDeviceRequestFailed(requests=requests)

    def __iter__(self):
        # 'devices' shouldn't be part of stats
        pools = []
        for pool in self.pools:
            tmp = {k: v for k, v in pool.items() if k != 'devices'}
            pools.append(tmp)
        return iter(pools)

    def clear(self):
        """Clear all the stats maintained."""
        self.pools = []

    def __eq__(self, other):
        return self.pools == other.pools

    if six.PY2:
        def __ne__(self, other):
            return not (self == other)

    def to_device_pools_obj(self):
        """Return the contents of the pools as a PciDevicePoolList object."""
        stats = [x for x in self]
        return pci_device_pool.from_pci_stats(stats)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils

from zun.common import exception
import zun.conf
from zun.db import api as dbapi
from zun.tests.unit.db import base
from zun.tests.unit.db import utils

CONF = zun.conf.CONF


class DbInventoryTestCase(base.DbTestCase):

    def setUp(self):
        super(DbInventoryTestCase, self).setUp()

    def test_create_inventory(self):
        utils.create_test_inventory(context=self.context)

    def test_create_inventory_already_exists(self):
        utils.create_test_inventory(
            context=self.context, resource_provider_id=1, resource_class_id=1)
        fields = {'resource_provider_id': 1, 'resource_class_id': 1}
        with self.assertRaisesRegex(exception.UniqueConstraintViolated,
                                    'A resource with %s *' % fields):
            utils.create_test_inventory(
                context=self.context, resource_provider_id=1,
                resource_class_id=1)

    def test_get_inventory_by_id(self):
        provider = utils.create_test_resource_provider(
            context=self.context)
        inventory = utils.create_test_inventory(
            resource_provider_id=provider.id, context=self.context)
        res = dbapi.get_inventory(self.context, inventory.id)
        self.assertEqual(inventory.id, res.id)

    def test_get_inventory_that_does_not_exist(self):
        inventory_id = 1111111
        self.assertRaises(exception.InventoryNotFound,
                          dbapi.get_inventory,
                          self.context,
                          inventory_id)

    def test_list_inventories(self):
        totals = []
        for i in range(1, 6):
            provider = utils.create_test_resource_provider(
                id=i,
                uuid=uuidutils.generate_uuid(),
                context=self.context)
            inventory = utils.create_test_inventory(
                id=i,
                resource_provider_id=provider.id,
                total=i,
                context=self.context)
            totals.append(inventory['total'])
        res = dbapi.list_inventories(self.context)
        res_totals = [r.total for r in res]
        self.assertEqual(sorted(totals), sorted(res_totals))

    def test_list_inventories_sorted(self):
        totals = []
        for i in range(5):
            provider = utils.create_test_resource_provider(
                id=i,
                uuid=uuidutils.generate_uuid(),
                context=self.context)
            inventory = utils.create_test_inventory(
                id=i,
                resource_provider_id=provider.id,
                total=10 - i,
                context=self.context)
            totals.append(inventory['total'])
        res = dbapi.list_inventories(self.context,
                                     sort_key='total')
        res_totals = [r.total for r in res]
        self.assertEqual(sorted(totals), res_totals)

        self.assertRaises(exception.InvalidParameterValue,
                          dbapi.list_inventories,
                          self.context,
                          sort_key='foo')

    def test_list_inventories_with_filters(self):
        provider = utils.create_test_resource_provider(
            context=self.context)
        inventory1 = utils.create_test_inventory(
            total=10,
            resource_provider_id=provider.id,
            resource_class_id=1,
            context=self.context)
        inventory2 = utils.create_test_inventory(
            total=20,
            resource_provider_id=provider.id,
            resource_class_id=2,
            context=self.context)

        res = dbapi.list_inventories(
            self.context, filters={'total': 10})
        self.assertEqual([inventory1.id], [r.id for r in res])

        res = dbapi.list_inventories(
            self.context, filters={'total': 20})
        self.assertEqual([inventory2.id], [r.id for r in res])

        res = dbapi.list_inventories(
            self.context, filters={'total': 11111})
        self.assertEqual([], [r.id for r in res])

        res = dbapi.list_inventories(
            self.context,
            filters={'total': inventory1.total})
        self.assertEqual([inventory1.id], [r.id for r in res])

    def test_destroy_inventory(self):
        inventory = utils.create_test_inventory(context=self.context)
        dbapi.destroy_inventory(self.context, inventory.id)
        self.assertRaises(exception.InventoryNotFound,
                          dbapi.get_inventory,
                          self.context, inventory.id)

    def test_destroy_inventory_by_id(self):
        inventory = utils.create_test_inventory(context=self.context)
        dbapi.destroy_inventory(self.context, inventory.id)
        self.assertRaises(exception.InventoryNotFound,
                          dbapi.get_inventory,
                          self.context, inventory.id)

    def test_destroy_inventory_that_does_not_exist(self):
        inventory_id = 1111111
        self.assertRaises(exception.InventoryNotFound,
                          dbapi.destroy_inventory, self.context,
                          inventory_id)

    def test_update_inventory(self):
        inventory = utils.create_test_inventory(context=self.context)
        old_total = inventory.total
        new_total = 100
        self.assertNotEqual(old_total, new_total)

        res = dbapi.update_inventory(self.context, inventory.id,
                                     {'total': new_total})
        self.assertEqual(new_total, res.total)

    def test_update_inventory_not_found(self):
        inventory_id = 11111111
        new_total = 'new-total'
        self.assertRaises(exception.InventoryNotFound,
                          dbapi.update_inventory, self.context,
                          inventory_id, {'total': new_total})



# -*- encoding: utf-8 -*-
#
# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The Zun Service API."""

import sys

from zun.common import profiler
from zun.common import service as zun_service
import zun.conf

CONF = zun.conf.CONF


def main():
    # Parse config file and command line options, then start logging
    zun_service.prepare_service(sys.argv)

    # Enable object backporting via the conductor
    # TODO(yuanying): Uncomment after rpc services are implemented
    # base.zunObject.indirection_api = base.zunObjectIndirectionAPI()

    # Setup OSprofiler for WSGI service
    profiler.setup('zun-api', CONF.api.host_ip)

    # Build and start the WSGI app
    launcher = zun_service.process_launcher()
    server = zun_service.WSGIService(
        'zun_api',
        CONF.api.enable_ssl_api
    )
    launcher.launch_service(server, workers=server.workers)
    launcher.wait()

if __name__ == '__main__':
    sys.exit(main())



# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy
# of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_utils import importutils


profiler_opts = importutils.try_import('osprofiler.opts')


def register_opts(conf):
    if profiler_opts:
        profiler_opts.set_defaults(conf)


def list_opts():
    if not profiler_opts:
        return {}
    return {
        profiler_opts._profiler_opt_group: profiler_opts._PROFILER_OPTS
    }



#    Copyright 2016 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import shlex
import sys

from oslo_log import log as logging
from oslo_privsep import priv_context
from oslo_service import service

from zun.common import rpc_service
from zun.common import service as zun_service
from zun.common import utils
import zun.conf

CONF = zun.conf.CONF
LOG = logging.getLogger(__name__)


def main():
    priv_context.init(root_helper=shlex.split(utils.get_root_helper()))
    zun_service.prepare_service(sys.argv)

    LOG.info('Starting server in PID %s', os.getpid())
    CONF.log_opt_values(LOG, logging.DEBUG)

    CONF.import_opt('topic', 'zun.conf.compute', group='compute')

    from zun.compute import manager as compute_manager
    endpoints = [
        compute_manager.Manager(),
    ]

    server = rpc_service.Service.create(CONF.compute.topic, CONF.host,
                                        endpoints, binary='zun-compute')
    launcher = service.launch(CONF, server, restart_method='mutate')
    launcher.wait()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_utils import strutils
import pecan

from zun.api.controllers import base
from zun.api.controllers import link
from zun.api.controllers.v1 import collection
from zun.api.controllers.v1.schemas import images as schema
from zun.api.controllers.v1.views import images_view as view
from zun.api import utils as api_utils
from zun.api import validation
from zun.common import exception
from zun.common.i18n import _
from zun.common import policy
from zun.common import utils
import zun.conf
from zun import objects

CONF = zun.conf.CONF
LOG = logging.getLogger(__name__)


def check_policy_on_image(image, action):
    context = pecan.request.context
    policy.enforce(context, action, image, action=action)


def _get_host(host_ident):
    try:
        return api_utils.get_resource('ComputeNode', host_ident)
    except exception.ComputeNodeNotFound:
        msg = _("The host %s does not exist.") % host_ident
        raise exception.InvalidValue(msg)


class ImageCollection(collection.Collection):
    """API representation of a collection of images."""

    fields = {
        'images',
        'next'
    }

    """A list containing images objects"""

    def __init__(self, **kwargs):
        super(ImageCollection, self).__init__(**kwargs)
        self._type = 'images'

    @staticmethod
    def convert_with_links(rpc_images, limit, url=None,
                           expand=False, **kwargs):
        collection = ImageCollection()
        collection.images = [view.format_image(url, p) for p in rpc_images]
        collection.next = collection.get_next(limit, url=url, **kwargs)
        return collection


class ImagesController(base.Controller):
    """Controller for Images"""

    _custom_actions = {
        'search': ['GET']
    }

    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    def delete(self, image_id):
        context = pecan.request.context
        policy.enforce(context, "image:delete", action="image:delete")
        image = utils.get_image(image_id)
        return pecan.request.compute_api.image_delete(context, image)

    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    def get_all(self, **kwargs):
        """Retrieve a list of images."""
        context = pecan.request.context
        policy.enforce(context, "image:get_all",
                       action="image:get_all")
        return self._get_images_collection(**kwargs)

    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    def get_one(self, image_id):
        """Retrieve information about the given image.

        :param image_id: UUID of a image.
        """
        image = utils.get_image(image_id)
        check_policy_on_image(image.as_dict(), "image:get_one")
        return view.format_image(pecan.request.host_url, image)

    def _get_images_collection(self, **kwargs):
        context = pecan.request.context
        limit = api_utils.validate_limit(kwargs.get('limit'))
        sort_dir = api_utils.validate_sort_dir(kwargs.get('sort_dir', 'asc'))
        sort_key = kwargs.get('sort_key', 'id')
        resource_url = kwargs.get('resource_url')
        expand = kwargs.get('expand')
        filters = None
        marker_obj = None
        marker = kwargs.get('marker')
        if marker:
            marker_obj = objects.Image.get_by_uuid(context, marker)
        images = objects.Image.list(context,
                                    limit,
                                    marker_obj,
                                    sort_key,
                                    sort_dir,
                                    filters=filters)
        return ImageCollection.convert_with_links(images, limit,
                                                  url=resource_url,
                                                  expand=expand,
                                                  sort_key=sort_key,
                                                  sort_dir=sort_dir)

    @pecan.expose('json')
    @api_utils.enforce_content_types(['application/json'])
    @exception.wrap_pecan_controller_exception
    @validation.validated(schema.image_create)
    def post(self, **image_dict):
        """Create a new image.

        :param image_dict: an image within the request body.
        """
        context = pecan.request.context
        policy.enforce(context, "image:pull",
                       action="image:pull")
        host = _get_host(image_dict.pop('host'))
        image_dict['project_id'] = context.project_id
        image_dict['user_id'] = context.user_id
        repo_tag = image_dict.get('repo')
        image_dict['host'] = host.hostname
        image_dict['repo'], image_dict['tag'] = utils.parse_image_name(
            repo_tag)
        new_image = objects.Image(context, **image_dict)
        new_image.pull(context)
        pecan.request.compute_api.image_pull(context, new_image)
        # Set the HTTP Location Header
        pecan.response.location = link.build_url('images', new_image.uuid)
        pecan.response.status = 202
        return view.format_image(pecan.request.host_url, new_image)

    @pecan.expose('json')
    @exception.wrap_pecan_controller_exception
    @validation.validate_query_param(pecan.request, schema.query_param_search)
    def search(self, image, image_driver=None, exact_match=False):
        """Search a specific image

        :param image:  Name of the image.
        :param image_driver: Name of the image driver (glance, docker).
        :param exact_match: if True, exact match the image name.
        """
        context = pecan.request.context
        policy.enforce(context, "image:search",
                       action="image:search")
        LOG.debug('Calling compute.image_search with %s', image)
        try:
            exact_match = strutils.bool_from_string(exact_match, strict=True)
        except ValueError:
            bools = ', '.join(strutils.TRUE_STRINGS + strutils.FALSE_STRINGS)
            raise exception.InvalidValue(_('Valid exact_match values are: %s')
                                         % bools)
        # Valiadtion accepts 'None' so need to convert it to None
        image_driver = api_utils.string_or_none(image_driver)
        if not image_driver:
            image_driver = CONF.default_image_driver

        return pecan.request.compute_api.image_search(context, image,
                                                      image_driver,
                                                      exact_match, None)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""drop foreign key of container_actions container_uuid

Revision ID: 8b0082d9e7c1
Revises: b6bfca998431
Create Date: 2017-12-14 15:32:38.520594

"""

# revision identifiers, used by Alembic.
revision = '8b0082d9e7c1'
down_revision = 'b6bfca998431'
branch_labels = None
depends_on = None

from alembic import op
from sqlalchemy.engine.reflection import Inspector as insp

CONTAINER_ACTIONS = 'container_actions'
CONTAINER = 'container'


def upgrade():
    bind = op.get_bind()

    inspector = insp.from_engine(bind)
    foreign_keys = inspector.get_foreign_keys(CONTAINER_ACTIONS)

    for foreign_key in foreign_keys:
        if foreign_key.get('referred_table') == CONTAINER:
            op.drop_constraint(foreign_key.get('name'), CONTAINER_ACTIONS,
                               type_="foreignkey")



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Tests For kuryr network
"""
import copy
import mock

from neutronclient.common import exceptions as n_exc

from zun.common import exception
from zun.network import kuryr_network
from zun.objects.container import Container
from zun.objects.network import Network
from zun.tests import base
from zun.tests.unit.db import utils


class FakeNeutronClient(object):

    def __init__(self):
        super(FakeNeutronClient, self).__init__()
        self.networks = [{'id': 'fake-net-id',
                          'name': 'fake-net-name',
                          'tenant_id': 'fake_project',
                          'shared': False}]
        self.subnets = [{'ip_version': 4,
                         'subnetpool_id': 'fake-subnetpool-id',
                         'cidr': '10.5.0.0/16',
                         'gateway_ip': '10.5.0.1',
                         'tenant_id': 'fake_project',
                         'network_id': 'fake-net-id',
                         'id': 'fake-subnet-id'}]
        self.ports = [{'id': 'fake-port-id',
                       'fixed_ips': [{
                           'ip_address': '10.5.0.22',
                           'subnet_id': 'fake-subnet-id'}],
                       'device_id': '',
                       'tenant_id': 'fake_project',
                       'security_groups': ['sg1']}]
        self.subnetpools = [{'address_scope_id': None,
                             'default_prefixlen': 8,
                             'id': 'fake-subnetpool-id',
                             'ip_version': 4,
                             'prefixes': ['10.5.0.0/16'],
                             'tags': ['06be906d-fe20-41af-b0e4-d770e185e7c8'],
                             'tenant_id': 'fake_project',
                             'name': 'test_subnet_pool-'}]

    def list_subnets(self, **filters):
        subnets = list(self.subnets)
        for subnet in self.subnets:
            for attr in filters:
                if subnet[attr] != filters[attr]:
                    subnets.remove(subnet)
        return {'subnets': copy.deepcopy(subnets)}

    def create_port(self, port):
        port_data = copy.deepcopy(port['port'])
        self.ports.append(port_data)
        return port

    def update_port(self, port_id, port, **kwargs):
        port_data = copy.deepcopy(port['port'])
        for port in self.ports:
            if port['id'] == port_id:
                port.update(port_data)

    def list_ports(self, **filters):
        ports = list(self.ports)
        for port in self.ports:
            for attr in filters:
                if port[attr] != filters[attr]:
                    ports.remove(port)
        return {'ports': copy.deepcopy(ports)}

    def delete_port(self, port_id):
        for port in self.ports:
            if port['id'] == port_id:
                self.ports.remove(port)
                return

    def get_neutron_port(self, port_id):
        for port in self.ports:
            if port['id'] == port_id:
                return copy.deepcopy(port)

    def get_neutron_network(self, network_id):
        for network in self.networks:
            if network['id'] == network_id or network['name'] == network_id:
                return copy.deepcopy(network)

    def list_subnetpools(self, **filters):
        subnetpools = list(self.subnetpools)
        for subnetpool in self.subnetpools:
            for attr in filters:
                if subnetpool[attr] != filters[attr]:
                    subnetpools.remove(subnetpool)
        return {'subnetpools': copy.deepcopy(subnetpools)}


class FakeDockerClient(object):

    def create_network(self, **kwargs):
        return {'Warning': '',
                'Id': '7372099cdcecbc9918d3666440b73a170d9690'}

    def remove_network(self, network_name):
        pass

    def inspect_network(self, network_name):
        return {'Name': 'c02afe4e-8350-4263-8078',
                'Options': {'neutron.net.uuid': '1234567'}}

    def networks(self, **kwargs):
        return [{'Name': 'test_network'}]

    def connect_container_to_network(self, container_id, network_name,
                                     **kwargs):
        pass

    def disconnect_container_from_network(self, container_id,
                                          network_name):
        pass


class KuryrNetworkTestCase(base.TestCase):
    """Test casse for kuryr network"""

    def setUp(self):
        super(KuryrNetworkTestCase, self).setUp()
        self.docker_api = FakeDockerClient()
        self.network_api = kuryr_network.KuryrNetwork()
        self.network_api.init(self.context, self.docker_api)
        self.network_api.neutron_api = FakeNeutronClient()

    @mock.patch.object(Network, 'create')
    @mock.patch.object(Network, 'save')
    @mock.patch('zun.network.neutron.NeutronAPI')
    def test_create_network_without_subnetpool(
            self, mock_neutron_api_cls, mock_save, mock_create):
        self.network_api.neutron_api.subnets[0].pop('subnetpool_id')
        mock_neutron_api_cls.return_value = self.network_api.neutron_api
        name = 'test_kuryr_network'
        neutron_net_id = 'fake-net-id'
        with mock.patch.object(self.network_api.docker, 'create_network',
                               return_value={'Id': 'docker-net'}
                               ) as mock_create_network:
            network = self.network_api.create_network(name, neutron_net_id)
        self.assertEqual('docker-net', network.network_id)
        mock_create_network.assert_called_once_with(
            name=name,
            driver='kuryr',
            enable_ipv6=False,
            ipam={'Config': [{'Subnet': '10.5.0.0/16', 'Gateway': '10.5.0.1'}],
                  'Driver': 'kuryr',
                  'Options': {'neutron.net.shared': 'False',
                              'neutron.subnet.uuid': 'fake-subnet-id'}},
            options={'neutron.net.uuid': 'fake-net-id',
                     'neutron.net.shared': 'False',
                     'neutron.subnet.uuid': 'fake-subnet-id'})

    @mock.patch.object(Network, 'create')
    @mock.patch.object(Network, 'save')
    @mock.patch('zun.network.neutron.NeutronAPI')
    def test_create_network_with_subnetpool(
            self, mock_neutron_api_cls, mock_save, mock_create):
        mock_neutron_api_cls.return_value = self.network_api.neutron_api
        name = 'test_kuryr_network'
        neutron_net_id = 'fake-net-id'
        with mock.patch.object(self.network_api.docker, 'create_network',
                               return_value={'Id': 'docker-net'}
                               ) as mock_create_network:
            network = self.network_api.create_network(name, neutron_net_id)
        self.assertEqual('docker-net', network.network_id)
        mock_create_network.assert_called_once_with(
            name=name,
            driver='kuryr',
            enable_ipv6=False,
            ipam={'Config': [{'Subnet': '10.5.0.0/16', 'Gateway': '10.5.0.1'}],
                  'Driver': 'kuryr',
                  'Options': {'neutron.net.shared': 'False',
                              'neutron.subnet.uuid': 'fake-subnet-id'}},
            options={'neutron.net.uuid': 'fake-net-id',
                     'neutron.net.shared': 'False',
                     'neutron.subnet.uuid': 'fake-subnet-id'})

    @mock.patch.object(Network, 'create')
    @mock.patch.object(Network, 'save')
    @mock.patch.object(Network, 'list')
    @mock.patch('zun.network.neutron.NeutronAPI')
    def test_create_network_already_exist(
            self, mock_neutron_api_cls, mock_list, mock_save, mock_create):
        mock_neutron_api_cls.return_value = self.network_api.neutron_api
        name = 'test_kuryr_network'
        neutron_net_id = 'fake-net-id'
        docker_net_id = 'docker-net'
        fake_network = mock.Mock()
        fake_network.network_id = docker_net_id
        mock_list.return_value = [fake_network]
        mock_create.side_effect = exception.NetworkAlreadyExists(
            field='neutron_net_id', value=neutron_net_id)
        with mock.patch.object(self.network_api.docker, 'networks',
                               return_value=[{'Id': docker_net_id}]
                               ) as mock_list_network:
            network = self.network_api.create_network(name, neutron_net_id)
        self.assertEqual(docker_net_id, network.network_id)
        mock_list.assert_called_once_with(
            self.context, filters={'neutron_net_id': neutron_net_id})
        mock_list_network.assert_called_once_with(names=[name])

    def test_remove_network(self):
        network = mock.Mock(name='c02afe4e-8350-4263-8078')
        self.network_api.remove_network(network)
        network.destroy.assert_called_once_with()

    def test_inspect_network(self):
        network_name = 'c02afe4e-8350-4263-8078'
        expected = {'Name': 'c02afe4e-8350-4263-8078',
                    'Options': {'neutron.net.uuid': '1234567'}}
        info = self.network_api.inspect_network(network_name)
        self.assertEqual(expected, info)

    def test_list_networks(self):
        expected = [{'Name': 'test_network'}]
        networks = self.network_api.list_networks()
        self.assertEqual(expected, networks)

    def test_connect_container_to_network(self):
        container = Container(self.context, **utils.get_test_container())
        network_name = 'c02afe4e-8350-4263-8078'
        requested_net = {'ipv4_address': '10.5.0.22',
                         'port': 'fake-port-id',
                         'preserve_on_delete': True}
        expected_address = [{'version': 4, 'addr': '10.5.0.22',
                             'port': 'fake-port-id',
                             'subnet_id': 'fake-subnet-id',
                             'preserve_on_delete': True}]
        old_port = self.network_api.neutron_api.list_ports(
            id='fake-port-id')['ports'][0]
        self.assertEqual('', old_port['device_id'])
        with mock.patch.object(self.network_api.docker,
                               'connect_container_to_network') as mock_connect:
            address = self.network_api.connect_container_to_network(
                container, network_name, requested_net)

        self.assertEqual(expected_address, address)
        mock_connect.assert_called_once_with(
            container.container_id, network_name, ipv4_address='10.5.0.22')
        new_port = self.network_api.neutron_api.list_ports(
            id='fake-port-id')['ports'][0]
        self.assertEqual(container.uuid, new_port['device_id'])

    @mock.patch('zun.network.neutron.NeutronAPI')
    def test_connect_container_to_network_failed(self, mock_neutron_api_cls):
        container = Container(self.context, **utils.get_test_container())
        network_name = 'c02afe4e-8350-4263-8078'
        requested_net = {'ipv4_address': '10.5.0.22',
                         'port': 'fake-port-id',
                         'preserve_on_delete': True}
        mock_neutron_api_cls.return_value = self.network_api.neutron_api
        old_port = self.network_api.neutron_api.list_ports(
            id='fake-port-id')['ports'][0]
        self.assertEqual('', old_port['device_id'])
        self.network_api.docker = mock.MagicMock()
        self.network_api.docker.connect_container_to_network = \
            mock.Mock(side_effect=exception.DockerError)
        self.assertRaises(exception.DockerError,
                          self.network_api.connect_container_to_network,
                          container, network_name, requested_net)
        new_port = self.network_api.neutron_api.list_ports(
            id='fake-port-id')['ports'][0]
        self.assertEqual('', new_port['device_id'])

    def test_disconnect_container_from_network(self):
        addresses = {'fake-net-id': [{'port': 'fake-port-id',
                                      'preserve_on_delete': False}]}
        container = Container(self.context, **utils.get_test_container(
            addresses=addresses))
        network_name = 'c02afe4e-8350-4263-8078'
        ports = self.network_api.neutron_api.list_ports(
            id='fake-port-id')['ports']
        self.assertEqual(1, len(ports))
        with mock.patch.object(self.network_api.docker,
                               'disconnect_container_from_network'
                               ) as mock_disconnect:
            self.network_api.disconnect_container_from_network(
                container, network_name, 'fake-net-id')
        mock_disconnect.assert_called_once_with(
            container.container_id, network_name)
        # assert the neutron port is deleted
        ports = self.network_api.neutron_api.list_ports(
            id='fake-port-id')['ports']
        self.assertEqual(0, len(ports))

    @mock.patch('zun.network.neutron.NeutronAPI')
    def test_add_security_groups_to_ports(self, mock_neutron_api_cls):
        addresses = {'fake-net-id': [{'port': 'fake-port-id'}]}
        container = Container(self.context, **utils.get_test_container(
            addresses=addresses))
        mock_neutron_api_cls.return_value = self.network_api.neutron_api
        old_port = self.network_api.neutron_api.list_ports(
            id='fake-port-id')['ports'][0]
        security_group_ids = ['sg2']
        self.network_api.add_security_groups_to_ports(container,
                                                      security_group_ids)
        new_port = self.network_api.neutron_api.list_ports(
            id='fake-port-id')['ports'][0]
        old_secgroups = old_port.pop('security_groups')
        new_secgroups = new_port.pop('security_groups')
        self.assertEqual(old_secgroups + ['sg2'], new_secgroups)
        # assert nothing else changed besides security_groups
        self.assertEqual(old_port, new_port)

    @mock.patch('zun.network.neutron.NeutronAPI')
    def test_add_security_groups_to_ports_bad_update(
            self, mock_neutron_api_cls):
        addresses = {'fake-net-id': [{'port': 'fake-port-id'}]}
        container = Container(self.context, **utils.get_test_container(
            addresses=addresses))
        mock_neutron_api_cls.return_value = self.network_api.neutron_api
        security_group_ids = ['sg2']
        with mock.patch.object(self.network_api.neutron_api,
                               'update_port') as mock_update_port:
            mock_update_port.side_effect = n_exc.BadRequest(
                message='error')
            self.assertRaises(exception.SecurityGroupCannotBeApplied,
                              self.network_api.add_security_groups_to_ports,
                              container, security_group_ids)

        mock_update_port.assert_called_once_with(
            'fake-port-id',
            {'port': {'security_groups': ['sg1', 'sg2']}},
            admin=True)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for manipulating resource classes via the DB API"""
from oslo_utils import uuidutils
import six

from zun.common import exception
import zun.conf
from zun.db import api as dbapi
from zun.tests.unit.db import base
from zun.tests.unit.db import utils

CONF = zun.conf.CONF


class DbResourceClassTestCase(base.DbTestCase):

    def setUp(self):
        super(DbResourceClassTestCase, self).setUp()

    def test_create_resource_class(self):
        utils.create_test_resource_class(context=self.context)

    def test_create_resource_class_already_exists(self):
        utils.create_test_resource_class(
            context=self.context, uuid='123')
        with self.assertRaisesRegex(exception.ResourceClassAlreadyExists,
                                    'A resource class with uuid 123.*'):
            utils.create_test_resource_class(
                context=self.context, uuid='123')

    def test_get_resource_class_by_uuid(self):
        resource = utils.create_test_resource_class(context=self.context)
        res = dbapi.get_resource_class(self.context, resource.uuid)
        self.assertEqual(resource.uuid, res.uuid)
        self.assertEqual(resource.name, res.name)

    def test_get_resource_class_by_name(self):
        resource = utils.create_test_resource_class(context=self.context)
        res = dbapi.get_resource_class(self.context, resource.name)
        self.assertEqual(resource.id, res.id)
        self.assertEqual(resource.name, res.name)

    def test_get_resource_class_that_does_not_exist(self):
        self.assertRaises(exception.ResourceClassNotFound,
                          dbapi.get_resource_class,
                          self.context, uuidutils.generate_uuid())

    def test_list_resource_classes(self):
        names = []
        for i in range(1, 6):
            resource = utils.create_test_resource_class(
                context=self.context,
                uuid=uuidutils.generate_uuid(),
                name='class' + str(i))
            names.append(six.text_type(resource['name']))
        res = dbapi.list_resource_classes(self.context)
        res_names = [r.name for r in res]
        self.assertEqual(sorted(names), sorted(res_names))

    def test_list_resource_classes_sorted(self):
        names = []
        for i in range(5):
            resource = utils.create_test_resource_class(
                context=self.context,
                uuid=uuidutils.generate_uuid(),
                name='class' + str(i))
            names.append(six.text_type(resource.name))
        res = dbapi.list_resource_classes(self.context, sort_key='name')
        res_names = [r.name for r in res]
        self.assertEqual(sorted(names), res_names)

        self.assertRaises(exception.InvalidParameterValue,
                          dbapi.list_resource_classes,
                          self.context,
                          sort_key='foo')

    def test_destroy_resource_class(self):
        resource = utils.create_test_resource_class(context=self.context)
        dbapi.destroy_resource_class(self.context, resource.id)
        self.assertRaises(exception.ResourceClassNotFound,
                          dbapi.get_resource_class,
                          self.context, resource.id)

    def test_destroy_resource_class_that_does_not_exist(self):
        bad_id = 1111111
        self.assertRaises(exception.ResourceClassNotFound,
                          dbapi.destroy_resource_class, self.context,
                          bad_id)

    def test_update_resource_class(self):
        resource = utils.create_test_resource_class(context=self.context)
        old_name = resource.name
        new_name = 'new-name'
        self.assertNotEqual(old_name, new_name)

        res = dbapi.update_resource_class(
            self.context, resource.id, {'name': new_name})
        self.assertEqual(new_name, res.name)

    def test_update_resource_class_not_found(self):
        bad_id = 1111111
        new_name = 'new-name'
        self.assertRaises(exception.ResourceClassNotFound,
                          dbapi.update_resource_class, self.context,
                          bad_id, {'name': new_name})



# Copyright 2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
from oslo_utils import uuidutils
import pecan

from zun.api.controllers import versions
from zun.common import exception
from zun.common.i18n import _
import zun.conf
from zun import objects

CONF = zun.conf.CONF


DOCKER_MINIMUM_MEMORY = 4 * 1024 * 1024


def string_or_none(value):
    if value in [None, 'None']:
        return None
    else:
        return value


def validate_limit(limit):
    try:
        if limit is not None and int(limit) <= 0:
            raise exception.InvalidValue(_("Limit must be positive integer"))
    except ValueError:
        raise exception.InvalidValue(_("Limit must be positive integer"))

    if limit is not None:
        return min(CONF.api.max_limit, int(limit))
    else:
        return CONF.api.max_limit


def validate_sort_dir(sort_dir):
    if sort_dir not in ['asc', 'desc']:
        raise exception.InvalidValue(_("Invalid sort direction: %s. "
                                       "Acceptable values are "
                                       "'asc' or 'desc'") % sort_dir)
    return sort_dir


def get_resource(resource, resource_ident):
    """Get the resource from the uuid or logical name.

    :param resource: the resource type.
    :param resource_ident: the UUID or logical name of the resource.

    :returns: The resource.
    """
    resource = getattr(objects, resource)
    context = pecan.request.context
    if context.is_admin:
        context.all_projects = True
    if uuidutils.is_uuid_like(resource_ident):
        return resource.get_by_uuid(context, resource_ident)

    return resource.get_by_name(context, resource_ident)


def _do_enforce_content_types(pecan_req, valid_content_types):
    """Content type enforcement

    Check to see that content type in the request is one of the valid
    types passed in by our caller.
    """
    if pecan_req.content_type not in valid_content_types:
        m = (
            "Unexpected content type: {type}. Expected content types "
            "are: {expected}"
        ).format(
            type=pecan_req.content_type.decode('utf-8'),
            expected=valid_content_types
        )
        pecan.abort(415, m)


def enforce_content_types(valid_content_types):
    """Decorator handling content type enforcement on behalf of REST verbs."""

    def content_types_decorator(fn):

        @functools.wraps(fn)
        def content_types_enforcer(inst, *args, **kwargs):
            _do_enforce_content_types(pecan.request, valid_content_types)
            return fn(inst, *args, **kwargs)

        return content_types_enforcer

    return content_types_decorator


def version_check(action, version):
    """Check whether the current version supports the operation.

    :param action: Operations to be executed.
    :param version: The minimum version required to perform the operation.

    """
    req_version = pecan.request.version
    min_version = versions.Version('', '', '', version)
    if req_version < min_version:
        raise exception.InvalidParamInVersion(param=action,
                                              req_version=req_version,
                                              min_version=min_version)



# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""empty message

Revision ID: 3298c6a5c3d9
Revises: 271c7f45982d
Create Date: 2018-04-28 06:32:26.493248

"""

# revision identifiers, used by Alembic.
revision = '3298c6a5c3d9'
down_revision = '271c7f45982d'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'network',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('project_id', sa.String(length=255), nullable=True),
        sa.Column('user_id', sa.String(length=255), nullable=True),
        sa.Column('name', sa.String(length=255), nullable=True),
        sa.Column('uuid', sa.String(length=36), nullable=True),
        sa.Column('network_id', sa.String(length=255), nullable=True),
        sa.Column('neutron_net_id', sa.String(length=255), nullable=True),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('uuid', name='uniq_network0uuid'),
        sa.UniqueConstraint('neutron_net_id',
                            name='uniq_network0neutron_net_id'),
        mysql_charset='utf8',
        mysql_engine='InnoDB'
        )



# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg


scheduler_group = cfg.OptGroup(name="scheduler",
                               title="Scheduler configuration")

scheduler_opts = [
    cfg.StrOpt("driver",
               default="filter_scheduler",
               choices=("chance_scheduler", "fake_scheduler",
                        "filter_scheduler"),
               help="""
The class of the driver used by the scheduler.

The options are chosen from the entry points under the namespace
'zun.scheduler.driver' in 'setup.cfg'.

Possible values:

* A string, where the string corresponds to the class name of a scheduler
  driver. There are a number of options available:
** 'chance_scheduler', which simply picks a host at random
** A custom scheduler driver. In this case, you will be responsible for
   creating and maintaining the entry point in your 'setup.cfg' file
"""),
    cfg.MultiStrOpt("available_filters",
                    default=["zun.scheduler.filters.all_filters"],
                    help="""
Filters that the scheduler can use.

An unordered list of the filter classes the zun scheduler may apply.  Only the
filters specified in the 'scheduler_enabled_filters' option will be used, but
any filter appearing in that option must also be included in this list.

By default, this is set to all filters that are included with zun.

This option is only used by the FilterScheduler and its subclasses; if you use
a different scheduler, this option has no effect.

Possible values:

* A list of zero or more strings, where each string corresponds to the name of
  a filter that may be used for selecting a host

Related options:

* scheduler_enabled_filters
"""),
    cfg.ListOpt("enabled_filters",
                default=[
                    "AvailabilityZoneFilter",
                    "CPUFilter",
                    "RamFilter",
                    "ComputeFilter",
                    "DiskFilter",
                    "RuntimeFilter",
                    ],
                help="""
Filters that the scheduler will use.

An ordered list of filter class names that will be used for filtering
hosts. Ignore the word 'default' in the name of this option: these filters will
*always* be applied, and they will be applied in the order they are listed so
place your most restrictive filters first to make the filtering process more
efficient.

This option is only used by the FilterScheduler and its subclasses; if you use
a different scheduler, this option has no effect.

Possible values:

* A list of zero or more strings, where each string corresponds to the name of
  a filter to be used for selecting a host

Related options:

* All of the filters in this option *must* be present in the
  'scheduler_available_filters' option, or a SchedulerHostFilterNotFound
  exception will be raised.
"""),
    cfg.IntOpt("max_placement_results",
               default=1000,
               min=1,
               help="""
This setting determines the maximum limit on results received from the
placement service during a scheduling operation. It effectively limits
the number of hosts that may be considered for scheduling requests that
match a large number of candidates.

A value of 1 (the minimum) will effectively defer scheduling to the placement
service strictly on "will it fit" grounds. A higher value will put an upper
cap on the number of results the scheduler will consider during the filtering
and weighing process. Large deployments may need to set this lower than the
total number of hosts available to limit memory consumption, network traffic,
etc. of the scheduler.

This option is only used by the FilterScheduler; if you use a different
scheduler, this option has no effect.
"""),
]


def register_opts(conf):
    conf.register_group(scheduler_group)
    conf.register_opts(scheduler_opts, group=scheduler_group)


def list_opts():
    return {scheduler_group: scheduler_opts}




#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add security groups

Revision ID: 174cafda0857
Revises: 5359d23b2322
Create Date: 2017-05-16 08:49:27.482284

"""

# revision identifiers, used by Alembic.
revision = '174cafda0857'
down_revision = '5359d23b2322'
branch_labels = None
depends_on = None


from alembic import op
import sqlalchemy as sa

import zun


def upgrade():
    op.add_column('container',
                  sa.Column('security_groups',
                            zun.db.sqlalchemy.models.JSONEncodedList(),
                            nullable=True))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import six

from stevedore import driver

import zun.conf


CONF = zun.conf.CONF


def api(*args, **kwargs):
    network_driver = CONF.network.driver
    network_api = driver.DriverManager(
        "zun.network.driver",
        network_driver,
        invoke_on_load=True).driver

    network_api.init(*args, **kwargs)
    return network_api


@six.add_metaclass(abc.ABCMeta)
class Network(object):
    """The base class that all Network classes should inherit from."""

    def init(self, context, *args, **kwargs):
        raise NotImplementedError()

    def create_network(self, *args, **kwargs):
        raise NotImplementedError()

    def remove_network(self, network_name, **kwargs):
        raise NotImplementedError()

    def inspect_network(self, network_name, **kwargs):
        raise NotImplementedError()

    def list_networks(self, **kwargs):
        raise NotImplementedError()

    def connect_container_to_network(self, container, network_name, **kwargs):
        raise NotImplementedError()

    def disconnect_container_from_network(self, container, network_name,
                                          **kwargs):
        raise NotImplementedError()

    def add_security_groups_to_ports(self, container, security_group_ids,
                                     **kwargs):
        raise NotImplementedError()

    def remove_security_groups_from_ports(self, container, security_group_ids,
                                          **kwargs):
        raise NotImplementedError()



# Copyright (c) 2013 dotCloud, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import multiprocessing
import os

from oslo_config import cfg
import psutil

CONF = cfg.CONF


def statvfs():
    docker_path = CONF.docker.root_directory
    if not os.path.exists(docker_path):
        docker_path = '/'
    return psutil.disk_usage(docker_path)


def get_disk_usage():
    # This is the location where Docker stores its containers. It's currently
    # hardcoded in Docker so it's not configurable yet.
    st = statvfs()
    return {
        'total': st.total,
        'available': st.free,
        'used': st.used
    }


def get_total_vcpus():
    return multiprocessing.cpu_count()


def get_vcpus_used(containers):
    total_vcpus_used = 0
    for container in containers:
        if isinstance(container, dict):
            total_vcpus_used += container.get('Config', {}).get(
                'CpuShares', 0) / 1024

    return total_vcpus_used


def get_memory_usage():
    vmem = psutil.virtual_memory()
    return {
        'total': vmem.total,
        'used': vmem.used
    }


def get_mounts():
    with open('/proc/mounts') as f:
        return f.readlines()


def get_cgroup_devices_path():
    for ln in get_mounts():
        fields = ln.split(' ')
        if fields[2] == 'cgroup' and 'devices' in fields[3].split(','):
            return fields[1]



# Copyright 2017 ARM Holdings.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from zun.api.controllers.v1.schemas import parameter_types

_capsule_properties = {
    'template': parameter_types.capsule_template
}

capsule_create = {
    'type': 'object',
    'properties': _capsule_properties,
    'required': ['template'],
    'additionalProperties': False
}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import mock

from oslo_config import cfg

from zun import objects
from zun.scheduler.client import query as scheduler_client
from zun.scheduler import filter_scheduler
from zun.tests import base
from zun.tests.unit.db import utils
from zun.tests.unit.scheduler import fakes


CONF = cfg.CONF


class SchedulerClientTestCase(base.TestCase):

    def setUp(self):
        super(SchedulerClientTestCase, self).setUp()
        self.mock_placement_client = mock.Mock()
        p = mock.patch('zun.scheduler.client.report.SchedulerReportClient',
                       return_value=self.mock_placement_client)
        p.start()
        self.addCleanup(p.stop)

        self.client_cls = scheduler_client.SchedulerClient
        self.client = self.client_cls()

    def test_init_using_default_schedulerdriver(self):
        driver = self.client_cls().driver
        self.assertIsInstance(driver, filter_scheduler.FilterScheduler)

    def test_init_using_custom_schedulerdriver(self):
        CONF.set_override('driver', 'fake_scheduler', group='scheduler')
        driver = self.client_cls().driver
        self.assertIsInstance(driver, fakes.FakeScheduler)

    @mock.patch('zun.scheduler.filter_scheduler.FilterScheduler'
                '.select_destinations')
    def test_select_destinations(self, mock_select_destinations):
        mock_alloc_req = {
            "allocations": {
                mock.sentinel.rp_uuid: [mock.sentinel.alloc_req]
            }
        }
        mock_provider_summaries = {
            mock.sentinel.rp_uuid: {}
        }
        self.mock_placement_client.get_allocation_candidates.return_value = (
            [mock_alloc_req], mock_provider_summaries,
            mock.sentinel.alloc_request_version
        )
        alloc_reqs_by_rp_uuid = collections.defaultdict(list)
        alloc_reqs_by_rp_uuid[mock.sentinel.rp_uuid] = [mock_alloc_req]

        containers = [objects.Container(self.context,
                                        **utils.get_test_container())]
        extra_spec = {}
        fake_args = ['ctxt', containers, extra_spec]
        self.client.select_destinations(*fake_args)
        mock_select_destinations.assert_called_once_with(
            'ctxt', containers, extra_spec, alloc_reqs_by_rp_uuid,
            mock_provider_summaries, mock.sentinel.alloc_request_version)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import mock
from oslo_config import cfg
import six

from zun.conf import opts
from zun.tests import base


class ConfTestCase(base.TestCase):

    def test_list_opts(self):
        for group, opt_list in opts.list_opts():
            if isinstance(group, six.string_types):
                self.assertEqual('DEFAULT', group)
            else:
                self.assertIsInstance(group, cfg.OptGroup)
            for opt in opt_list:
                self.assertIsInstance(opt, cfg.Opt)

    def test_list_module_name_invalid_mods(self):
        with mock.patch('pkgutil.iter_modules') as mock_mods:
            mock_mods.return_value = [(None, 'foo', True),
                                      (None, 'opts', False)]
            self.assertEqual([], opts._list_module_names())

    def test_list_module_name_valid_mods(self):
        with mock.patch('pkgutil.iter_modules') as mock_mods:
            mock_mods.return_value = [(None, 'foo', False)]
            self.assertEqual(['foo'], opts._list_module_names())

    def test_import_mods_no_func(self):
        modules = ['foo', 'bar']
        with mock.patch('importlib.import_module') as mock_import:
            mock_import.return_value = mock.sentinel.mods
            self.assertRaises(AttributeError, opts._import_modules, modules)
            mock_import.assert_called_once_with('zun.conf.foo')

    def test_import_mods_valid_func(self):
        modules = ['foo', 'bar']
        with mock.patch('importlib.import_module') as mock_import:
            mock_mod = mock.MagicMock()
            mock_import.return_value = mock_mod
            self.assertEqual([mock_mod, mock_mod],
                             opts._import_modules(modules))
            mock_import.assert_has_calls([mock.call('zun.conf.foo'),
                                          mock.call('zun.conf.bar')])

    def test_append_config(self):
        opt = collections.defaultdict(list)
        mock_module = mock.MagicMock()
        mock_conf = mock.MagicMock()
        mock_module.list_opts.return_value = mock_conf
        mock_conf.items.return_value = [('foo', 'bar')]
        opts._append_config_options([mock_module], opt)
        self.assertEqual({'foo': ['b', 'a', 'r']}, opt)



# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import mock

import zun
from zun.common import context
from zun.objects import fields
from zun.pci import manager
from zun.tests.unit.db import base
from zun.tests.unit.pci import fakes as pci_fakes
from zun.tests import uuidsentinel


fake_pci = {
    'compute_node_uuid': 1,
    'address': '0000:00:00.1',
    'product_id': 'p',
    'vendor_id': 'v',
    'request_id': None,
    'status': fields.PciDeviceStatus.AVAILABLE,
    'dev_type': fields.PciDeviceType.STANDARD,
    'parent_addr': None,
    'numa_node': 0}
fake_pci_1 = dict(fake_pci, address='0000:00:00.2',
                  product_id='p1', vendor_id='v1')
fake_pci_2 = dict(fake_pci, address='0000:00:00.3')

fake_pci_3 = dict(fake_pci, address='0000:00:01.1',
                  dev_type=fields.PciDeviceType.SRIOV_PF,
                  vendor_id='v2', product_id='p2', numa_node=None)
fake_pci_4 = dict(fake_pci, address='0000:00:02.1',
                  dev_type=fields.PciDeviceType.SRIOV_VF,
                  parent_addr='0000:00:01.1',
                  vendor_id='v2', product_id='p2', numa_node=None)
fake_pci_5 = dict(fake_pci, address='0000:00:02.2',
                  dev_type=fields.PciDeviceType.SRIOV_VF,
                  parent_addr='0000:00:01.1',
                  vendor_id='v2', product_id='p2', numa_node=None)

fake_db_dev = {
    'created_at': None,
    'updated_at': None,
    'deleted_at': None,
    'deleted': None,
    'id': 1,
    'uuid': uuidsentinel.pci_device1,
    'compute_node_uuid': 1,
    'address': '0000:00:00.1',
    'vendor_id': 'v',
    'product_id': 'p',
    'numa_node': 1,
    'dev_type': fields.PciDeviceType.STANDARD,
    'status': fields.PciDeviceStatus.AVAILABLE,
    'dev_id': 'i',
    'label': 'l',
    'container_uuid': None,
    'extra_info': '{}',
    'request_id': None,
    'parent_addr': None,
    }
fake_db_dev_1 = dict(fake_db_dev, vendor_id='v1',
                     uuid=uuidsentinel.pci_device1,
                     product_id='p1', id=2,
                     address='0000:00:00.2',
                     numa_node=0)
fake_db_dev_2 = dict(fake_db_dev, id=3, address='0000:00:00.3',
                     uuid=uuidsentinel.pci_device2,
                     numa_node=None, parent_addr='0000:00:00.1')
fake_db_devs = [fake_db_dev, fake_db_dev_1, fake_db_dev_2]

fake_db_dev_3 = dict(fake_db_dev, id=4, address='0000:00:01.1',
                     uuid=uuidsentinel.pci_device3,
                     vendor_id='v2', product_id='p2',
                     numa_node=None, dev_type=fields.PciDeviceType.SRIOV_PF)
fake_db_dev_4 = dict(fake_db_dev, id=5, address='0000:00:02.1',
                     uuid=uuidsentinel.pci_device4,
                     numa_node=None, dev_type=fields.PciDeviceType.SRIOV_VF,
                     vendor_id='v2', product_id='p2',
                     parent_addr='0000:00:01.1')
fake_db_dev_5 = dict(fake_db_dev, id=6, address='0000:00:02.2',
                     uuid=uuidsentinel.pci_device5,
                     numa_node=None, dev_type=fields.PciDeviceType.SRIOV_VF,
                     vendor_id='v2', product_id='p2',
                     parent_addr='0000:00:01.1')
fake_db_devs_tree = [fake_db_dev_3, fake_db_dev_4, fake_db_dev_5]


class PciDevTrackerTestCase(base.DbTestCase):
    def _fake_get_pci_devices(self, node_id):
                return self.fake_devs

    def _fake_pci_device_update(self, node_id, address, value):
        self.update_called += 1
        self.called_values = value
        fake_return = copy.deepcopy(fake_db_dev)
        return fake_return

    def _fake_pci_device_destroy(self, node_id, address):
        self.destroy_called += 1

    def _create_tracker(self, fake_devs):
        self.fake_devs = fake_devs
        self.tracker = manager.PciDevTracker(self.fake_context, 1)

    def setUp(self):
        super(PciDevTrackerTestCase, self).setUp()
        self.fake_context = context.get_admin_context()
        self.fake_devs = fake_db_devs[:]
        self.stub_out('zun.db.api.get_all_pci_device_by_node',
                      self._fake_get_pci_devices)
        # The fake_pci_whitelist must be called before creating the fake
        # devices
        patcher = pci_fakes.fake_pci_whitelist()
        self.addCleanup(patcher.stop)
        self._create_tracker(fake_db_devs[:])

    def test_pcidev_tracker_create(self):
        self.assertEqual(len(self.tracker.pci_devs), 3)
        free_devs = self.tracker.pci_stats.get_free_devs()
        self.assertEqual(len(free_devs), 3)
        self.assertEqual(list(self.tracker.stale), [])
        self.assertEqual(len(self.tracker.stats.pools), 3)
        self.assertEqual(self.tracker.node_id, 1)
        for dev in self.tracker.pci_devs:
            self.assertIsNone(dev.parent_device)
            self.assertEqual(dev.child_devices, [])

    def test_pcidev_tracker_create_device_tree(self):
        self._create_tracker(fake_db_devs_tree)

        self.assertEqual(len(self.tracker.pci_devs), 3)
        free_devs = self.tracker.pci_stats.get_free_devs()
        self.assertEqual(len(free_devs), 3)
        self.assertEqual(list(self.tracker.stale), [])
        self.assertEqual(len(self.tracker.stats.pools), 2)
        self.assertEqual(self.tracker.node_id, 1)
        pf = [dev for dev in self.tracker.pci_devs
              if dev.dev_type == fields.PciDeviceType.SRIOV_PF].pop()
        vfs = [dev for dev in self.tracker.pci_devs
               if dev.dev_type == fields.PciDeviceType.SRIOV_VF]
        self.assertEqual(2, len(vfs))

        # Assert we build the device tree correctly
        self.assertEqual(vfs, pf.child_devices)
        for vf in vfs:
            self.assertEqual(vf.parent_device, pf)

    def test_pcidev_tracker_create_device_tree_pf_only(self):
        self._create_tracker([fake_db_dev_3])

        self.assertEqual(len(self.tracker.pci_devs), 1)
        free_devs = self.tracker.pci_stats.get_free_devs()
        self.assertEqual(len(free_devs), 1)
        self.assertEqual(list(self.tracker.stale), [])
        self.assertEqual(len(self.tracker.stats.pools), 1)
        self.assertEqual(self.tracker.node_id, 1)
        pf = self.tracker.pci_devs[0]
        self.assertIsNone(pf.parent_device)
        self.assertEqual([], pf.child_devices)

    def test_pcidev_tracker_create_device_tree_vf_only(self):
        self._create_tracker([fake_db_dev_4])

        self.assertEqual(len(self.tracker.pci_devs), 1)
        free_devs = self.tracker.pci_stats.get_free_devs()
        self.assertEqual(len(free_devs), 1)
        self.assertEqual(list(self.tracker.stale), [])
        self.assertEqual(len(self.tracker.stats.pools), 1)
        self.assertEqual(self.tracker.node_id, 1)
        vf = self.tracker.pci_devs[0]
        self.assertIsNone(vf.parent_device)
        self.assertEqual([], vf.child_devices)

    @mock.patch.object(zun.objects.PciDevice, 'list_by_compute_node')
    def test_pcidev_tracker_create_no_nodeid(self, mock_get_cn):
        self.tracker = manager.PciDevTracker(self.fake_context)
        self.assertEqual(len(self.tracker.pci_devs), 0)
        self.assertFalse(mock_get_cn.called)

    @mock.patch.object(zun.objects.PciDevice, 'list_by_compute_node')
    def test_pcidev_tracker_create_with_nodeid(self, mock_get_cn):
        self.tracker = manager.PciDevTracker(self.fake_context, node_id=1)
        mock_get_cn.assert_called_once_with(self.fake_context, 1)

    def test_set_hvdev_new_dev(self):
        fake_pci_3 = dict(fake_pci, address='0000:00:00.4', vendor_id='v2')
        fake_pci_devs = [copy.deepcopy(fake_pci), copy.deepcopy(fake_pci_1),
                         copy.deepcopy(fake_pci_2), copy.deepcopy(fake_pci_3)]
        self.tracker._set_hvdevs(fake_pci_devs)
        self.assertEqual(len(self.tracker.pci_devs), 4)
        self.assertEqual(set([dev.address for
                              dev in self.tracker.pci_devs]),
                         set(['0000:00:00.1', '0000:00:00.2',
                              '0000:00:00.3', '0000:00:00.4']))
        self.assertEqual(set([dev.vendor_id for
                              dev in self.tracker.pci_devs]),
                         set(['v', 'v1', 'v2']))

    def test_set_hvdev_new_dev_tree_maintained(self):
        # Make sure the device tree is properly maintained when there are new
        # devices reported by the driver
        self._create_tracker(fake_db_devs_tree)

        fake_new_device = dict(fake_pci_5, id=12, address='0000:00:02.3')
        fake_pci_devs = [copy.deepcopy(fake_pci_3),
                         copy.deepcopy(fake_pci_4),
                         copy.deepcopy(fake_pci_5),
                         copy.deepcopy(fake_new_device)]
        self.tracker._set_hvdevs(fake_pci_devs)
        self.assertEqual(len(self.tracker.pci_devs), 4)

        pf = [dev for dev in self.tracker.pci_devs
              if dev.dev_type == fields.PciDeviceType.SRIOV_PF].pop()
        vfs = [dev for dev in self.tracker.pci_devs
               if dev.dev_type == fields.PciDeviceType.SRIOV_VF]
        self.assertEqual(3, len(vfs))

        # Assert we build the device tree correctly
        self.assertEqual(vfs, pf.child_devices)
        for vf in vfs:
            self.assertEqual(vf.parent_device, pf)

    def test_set_hvdev_changed(self):
        fake_pci_v2 = dict(fake_pci, address='0000:00:00.2', vendor_id='v1')
        fake_pci_devs = [copy.deepcopy(fake_pci), copy.deepcopy(fake_pci_2),
                         copy.deepcopy(fake_pci_v2)]
        self.tracker._set_hvdevs(fake_pci_devs)
        self.assertEqual(set([dev.vendor_id for
                             dev in self.tracker.pci_devs]),
                         set(['v', 'v1']))

    def test_set_hvdev_remove(self):
        self.tracker._set_hvdevs([fake_pci])
        self.assertEqual(
            len([dev for dev in self.tracker.pci_devs
                 if dev.status == fields.PciDeviceStatus.REMOVED]),
            2)

    def test_set_hvdev_remove_tree_maintained(self):
        # Make sure the device tree is properly maintained when there are
        # devices removed from the system (not reported by the driver but known
        # from previous scans)
        self._create_tracker(fake_db_devs_tree)

        fake_pci_devs = [copy.deepcopy(fake_pci_3), copy.deepcopy(fake_pci_4)]
        self.tracker._set_hvdevs(fake_pci_devs)
        self.assertEqual(
            2,
            len([dev for dev in self.tracker.pci_devs
                 if dev.status != fields.PciDeviceStatus.REMOVED]))
        pf = [dev for dev in self.tracker.pci_devs
              if dev.dev_type == fields.PciDeviceType.SRIOV_PF].pop()
        vfs = [dev for dev in self.tracker.pci_devs
               if (dev.dev_type == fields.PciDeviceType.SRIOV_VF and
                   dev.status != fields.PciDeviceStatus.REMOVED)]
        self.assertEqual(1, len(vfs))

        self.assertEqual(vfs, pf.child_devices)
        self.assertEqual(vfs[0].parent_device, pf)

    def test_save(self):
        self.stub_out('zun.db.api.update_pci_device',
                      self._fake_pci_device_update)
        fake_pci_v3 = dict(fake_pci, address='0000:00:00.2', vendor_id='v3')
        fake_pci_devs = [copy.deepcopy(fake_pci), copy.deepcopy(fake_pci_2),
                         copy.deepcopy(fake_pci_v3)]
        self.tracker._set_hvdevs(fake_pci_devs)
        self.update_called = 0
        self.tracker.save()
        self.assertEqual(self.update_called, 3)

    def test_save_removed(self):
        self.stub_out('zun.db.api.update_pci_device',
                      self._fake_pci_device_update)
        self.stub_out('zun.db.api.destroy_pci_device',
                      self._fake_pci_device_destroy)
        self.assertEqual(len(self.tracker.pci_devs), 3)
        dev = self.tracker.pci_devs[0]
        self.destroy_called = 0
        self.update_called = 0
        dev.remove()
        self.tracker.save()
        self.assertEqual(len(self.tracker.pci_devs), 2)
        self.assertEqual(self.destroy_called, 1)
